import traceback
from typing import List

from metricflow.errors.errors import ParsingException
from metricflow.model.objects.metric import Metric, MetricType, CumulativeMetricWindow
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.validations.validator_helpers import (
    ModelValidationRule,
    ValidationIssue,
    ValidationIssueType,
    ValidationError,
    ValidationFatal,
    ModelObjectType,
    validate_safely,
)


class MetricMeasuresRule(ModelValidationRule):
    """Checks that the measures referenced in the metrics exist."""

    @staticmethod
    @validate_safely(whats_being_done="checking measures referenced by the metric are exist")
    def _validate_metric_measure_references(
        metric: Metric, valid_measure_names: List[str]
    ) -> List[ValidationIssueType]:
        issues: List[ValidationIssueType] = []
        measures_in_metric = []

        if metric.type_params:
            if metric.type_params.measures:
                measures_in_metric.extend(metric.type_params.measures)
            if metric.type_params.numerator:
                measures_in_metric.append(metric.type_params.numerator)
            if metric.type_params.denominator:
                measures_in_metric.append(metric.type_params.denominator)

        for measure_in_metric in measures_in_metric:
            if measure_in_metric.element_name not in valid_measure_names:
                issues.append(
                    ValidationFatal(
                        model_object_reference=ValidationIssue.make_object_reference(
                            metric_name=metric.name,
                        ),
                        message=f"Invalid measure {measure_in_metric} in metric {metric.name}",
                    )
                )
        return issues

    @staticmethod
    @validate_safely(whats_being_done="running model validation ensuring metric measures exist")
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssueType]:  # noqa: D
        issues: List[ValidationIssueType] = []
        valid_measure_names = []
        for data_source in model.data_sources:
            for measure in data_source.measures:
                valid_measure_names.append(measure.name.element_name)

        for metric in model.metrics or []:
            issues += MetricMeasuresRule._validate_metric_measure_references(
                metric=metric, valid_measure_names=valid_measure_names
            )
        return issues


class CumulativeMetricRule(ModelValidationRule):
    """Checks that cumulative sum metrics are configured properly"""

    @staticmethod
    @validate_safely(whats_being_done="checking that the params of metric are valid if it is a cumulative sum metric")
    def _validate_cumulative_sum_metric_params(metric: Metric) -> List[ValidationIssueType]:
        issues: List[ValidationIssueType] = []

        if metric.type == MetricType.CUMULATIVE:
            if metric.type_params.window and metric.type_params.grain_to_date:
                issues.append(
                    ValidationError(
                        model_object_reference=ValidationIssue.make_object_reference(
                            object_type=ModelObjectType.METRIC,
                            object_name=metric.name,
                        ),
                        message="Both window and grain_to_date set for cumulative metric. Please set one or the other",
                    )
                )

            if metric.type_params.window:
                try:
                    _, _ = CumulativeMetricWindow.parse(metric.type_params.window.to_string())
                except ParsingException:
                    issues.append(
                        ValidationError(
                            model_object_reference=ValidationIssue.make_object_reference(
                                metric_name=metric.name,
                            ),
                            message=traceback.format_exc(),
                        )
                    )

        return issues

    @staticmethod
    @validate_safely(whats_being_done="running model validation ensuring cumulative sum metrics are valid")
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssueType]:  # noqa: D
        issues: List[ValidationIssueType] = []

        for metric in model.metrics or []:
            issues += CumulativeMetricRule._validate_cumulative_sum_metric_params(metric=metric)

        return issues



from typing import List, Any, Optional, Set

import ciso8601  # type: ignore
import pandas as pd
from sqlalchemy.engine import make_url

from metricflow.protocols.sql_client import SqlClient
from metricflow.sql_clients.big_query import BigQuerySqlClient
from metricflow.sql_clients.common_client import SqlDialect
from metricflow.sql_clients.redshift import RedshiftSqlClient
from metricflow.sql_clients.snowflake import SnowflakeSqlClient
from metricflow.sql_clients.sqlite import SqliteSqlClient


def make_df(  # type: ignore [misc]
    sql_client: SqlClient, columns: List[str], data: Any, time_columns: Optional[Set[str]] = None
) -> pd.DataFrame:
    """Helper to make a dataframe, converting the time columns to appropriate types."""
    time_columns = time_columns or set()
    # Should only be used in testing, so sql_client should be set.
    assert sql_client

    if sql_client.sql_engine_attributes.timestamp_type_supported:
        new_rows = []
        for row in data:
            new_row = []
            # Change the type of the column if it's in time_columns
            for i, column in enumerate(columns):
                if column in time_columns:
                    # ts_suffix = " 00:00:00" if ":" not in row[i] else ""
                    # ts_input = row[i] + ts_suffix
                    new_row.append(ciso8601.parse_datetime(row[i]))

                else:
                    new_row.append(row[i])
            new_rows.append(new_row)
        data = new_rows

    return pd.DataFrame(
        columns=columns,
        data=data,
    )


def make_sql_client(url: str, password: str) -> SqlClient:  # noqa: D
    dialect_protocol = make_url(url).drivername.split("+")
    dialect = SqlDialect(dialect_protocol[0])
    if len(dialect_protocol) > 2:
        raise ValueError(f"Invalid # of +'s in {url}")

    if dialect == SqlDialect.REDSHIFT:
        return RedshiftSqlClient.from_connection_details(url, password)
    elif dialect == SqlDialect.SNOWFLAKE:
        return SnowflakeSqlClient.from_connection_details(url, password)
    elif dialect == SqlDialect.BIGQUERY:
        return BigQuerySqlClient.from_connection_details(url, password)
    elif dialect == SqlDialect.SQLITE:
        return SqliteSqlClient.from_connection_details(url, password)
    else:
        raise ValueError(f"Unknown dialect: `{dialect}` in URL {url}")



import logging
from datetime import date
from typing import List, MutableSet

from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.elements.identifier import IdentifierType
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.validations.validator_helpers import (
    ModelValidationRule,
    ValidationIssue,
    ValidationError,
    ValidationIssueType,
    validate_safely,
)
from metricflow.model.validations.validator_helpers import ValidationFutureError

logger = logging.getLogger(__name__)


class IdentifierConfigRule(ModelValidationRule):
    """Checks that data source identifiers are valid"""

    @staticmethod
    @validate_safely(whats_being_done="running model validation ensuring identifiers are valid")
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssueType]:  # noqa: D
        issues = []
        for data_source in model.data_sources:
            issues += IdentifierConfigRule._validate_data_source_identifiers(data_source=data_source)
        return issues

    @staticmethod
    @validate_safely(whats_being_done="checking that the data source's identifiers are valid")
    def _validate_data_source_identifiers(data_source: DataSource) -> List[ValidationIssueType]:
        """Checks validity of composite identifiers"""
        issues: List[ValidationIssueType] = []
        for ident in data_source.identifiers:
            if ident.identifiers:
                for sub_id in ident.identifiers:
                    if sub_id.ref and (sub_id.name or sub_id.expr):
                        logger.warning(f"Identifier with error is: {ident}")
                        issues.append(
                            ValidationError(
                                model_object_reference=ValidationIssue.make_object_reference(
                                    data_source_name=data_source.name,
                                    identifier_name=ident.name.element_name,
                                ),
                                message=f"Both ref and name/expr set in sub identifier of identifier "
                                f"({ident.name.element_name}), please set one",
                            )
                        )
                    elif sub_id.ref is not None and sub_id.ref not in [i.name for i in data_source.identifiers]:
                        issues.append(
                            ValidationError(
                                model_object_reference=ValidationIssue.make_object_reference(
                                    data_source_name=data_source.name, identifier_name=ident.name.element_name
                                ),
                                message=f"Identifier ref must reference an existing identifier by name. "
                                f"No identifier in this data source has name: {sub_id.ref}",
                            )
                        )
                    elif not sub_id.ref and not sub_id.name:
                        issues.append(
                            ValidationError(
                                model_object_reference=ValidationIssue.make_object_reference(
                                    data_source_name=data_source.name, identifier_name=ident.name.element_name
                                ),
                                message=f"Must provide either name or ref for sub identifier of identifier "
                                f"with name: {ident.name.element_name}",
                            )
                        )

                    if sub_id.name:
                        for i in data_source.identifiers:
                            if i.name == sub_id.name and i.expr != sub_id.expr:
                                issues.append(
                                    ValidationError(
                                        model_object_reference=ValidationIssue.make_object_reference(
                                            data_source_name=data_source.name, identifier_name=ident.name.element_name
                                        ),
                                        message=f"If sub identifier has same name ({sub_id.name.element_name}) "
                                        f"as an existing Identifier they must have the same expr",
                                    )
                                )
                                break

        return issues


class OnePrimaryIdentifierPerDataSourceRule(ModelValidationRule):
    """Ensures that each data source has only one primary identifier"""

    @staticmethod
    @validate_safely(whats_being_done="checking data source has only one primary identifier")
    def _only_one_primary_identifier(data_source: DataSource) -> List[ValidationIssue]:
        primary_identifier_names: MutableSet[str] = set()
        for identifier in data_source.identifiers or []:
            if identifier.type == IdentifierType.PRIMARY:
                primary_identifier_names.add(identifier.name.element_name)

        if len(primary_identifier_names) > 1:
            return [
                ValidationFutureError(
                    model_object_reference=ValidationIssue.make_object_reference(data_source_name=data_source.name),
                    message=f"Data sources can have only one primary identifier. The data source"
                    f" `{data_source.name}` has {len(primary_identifier_names)}: {', '.join(primary_identifier_names)}",
                    error_date=date(2022, 1, 12),  # Wed January 12th 2022
                )
            ]
        return []

    @staticmethod
    @validate_safely(
        whats_being_done="running model validation ensuring each data source has only one primary identifier"
    )
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssue]:  # noqa: D
        issues = []

        for data_source in model.data_sources:
            issues += OnePrimaryIdentifierPerDataSourceRule._only_one_primary_identifier(data_source=data_source)

        return issues



from collections import OrderedDict
from typing import Set, Union

import pandas as pd
import pytest

from metricflow.dataflow.sql_table import SqlTable
from metricflow.object_utils import random_id
from metricflow.protocols.sql_client import SqlClient
from metricflow.sql.sql_bind_parameters import SqlBindParameters
from metricflow.sql_clients.sql_utils import make_df
from metricflow.test.compare_df import assert_dataframes_equal
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState


def _random_table() -> str:
    return f"test_table_{random_id()}"


def _select_x_as_y(sql_client: SqlClient, x: int = 1, y: str = "y") -> str:  # noqa: D
    return f"SELECT {x} AS {y}"


def _check_1col(df: pd.DataFrame, col: str = "y", vals: Set[Union[int, str]] = {1}) -> None:  # noqa: D
    assert isinstance(df, pd.DataFrame)
    assert df.shape == (len(vals), 1)
    assert df.columns.tolist() == [col]
    assert set(df[col]) == vals


def test_query(sql_client: SqlClient) -> None:  # noqa: D
    df = sql_client.query(_select_x_as_y(sql_client))
    _check_1col(df)


def test_query_with_execution_params(sql_client: SqlClient) -> None:  # noqa: D
    expr = "SELECT :x as y"
    sql_execution_params = SqlBindParameters()
    sql_execution_params.param_dict = OrderedDict([("x", "1")])
    df = sql_client.query(expr, sql_bind_parameters=sql_execution_params)
    assert isinstance(df, pd.DataFrame)
    assert df.shape == (1, 1)
    assert df.columns.tolist() == ["y"]
    assert set(df["y"]) == {"1"}


def test_select_one_query(sql_client: SqlClient) -> None:  # noqa: D
    sql_client.query("SELECT 1")
    with pytest.raises(Exception):
        sql_client.query("this is garbage")


def test_failed_query_with_execution_params(sql_client: SqlClient) -> None:  # noqa: D
    expr = "SELECT :x"
    sql_execution_params = SqlBindParameters()
    sql_execution_params.param_dict = OrderedDict([("x", "1")])

    sql_client.query(expr, sql_bind_parameters=sql_execution_params)
    with pytest.raises(Exception):
        sql_client.query("this is garbage")


def test_create_table(mf_test_session_state: MetricFlowTestSessionState, sql_client: SqlClient) -> None:  # noqa: D
    sql_table = SqlTable(schema_name=mf_test_session_state.mf_source_schema, table_name=_random_table())
    sql_client.create_table_as_select(sql_table, _select_x_as_y(sql_client))
    df = sql_client.query(f"SELECT * FROM {sql_table.sql}")
    _check_1col(df)


def test_create_table_from_dataframe(  # noqa: D
    mf_test_session_state: MetricFlowTestSessionState, sql_client: SqlClient
) -> None:
    expected_df = make_df(
        sql_client=sql_client,
        columns=["int_col", "str_col", "float_col", "bool_col", "time_col"],
        time_columns={"time_col"},
        data=[
            (1, "abc", 1.23, False, "2020-01-01"),
            (2, "def", 4.56, True, "2020-01-02"),
        ],
    )
    sql_table = SqlTable(schema_name=mf_test_session_state.mf_source_schema, table_name=_random_table())
    sql_client.create_table_from_dataframe(sql_table=sql_table, df=expected_df)

    actual_df = sql_client.query(f"SELECT * FROM {sql_table.sql}")
    assert_dataframes_equal(actual=actual_df, expected=expected_df)


def test_table_exists(mf_test_session_state: MetricFlowTestSessionState, sql_client: SqlClient) -> None:  # noqa: D
    sql_table = SqlTable(schema_name=mf_test_session_state.mf_source_schema, table_name=_random_table())
    sql_client.create_table_as_select(sql_table, _select_x_as_y(sql_client))
    assert sql_client.table_exists(sql_table)


def test_percent_signs_in_query(sql_client: SqlClient) -> None:  # noqa: D
    stmt = "SELECT foo FROM ( SELECT 'abba' AS foo ) WHERE foo LIKE '%a'"
    sql_client.query(stmt)
    df = sql_client.query(stmt)
    _check_1col(df, col="foo", vals={"abba"})


@pytest.fixture()
def example_df() -> pd.DataFrame:
    """Data frame containing data of different types for testing. DateTime would be good to add."""
    return pd.DataFrame(
        columns=["int_col", "str_col", "float_col", "bool_col"],
        data=[
            (1, "abc", 1.23, False),
            (2, "def", 4.56, True),
        ],
    )



from __future__ import annotations

import click
import datetime
import pandas as pd
import pytest

from click.testing import CliRunner, Result
from typing import List, Optional, Sequence

from metricflow.cli.cli_context import CLIContext
from metricflow.cli.models import Dimension, Materialization, Metric
from metricflow.engine.metricflow_engine import MetricFlowQueryRequest, MetricFlowQueryResult


class MockMetricFlowEngine:
    """Mock MetricFlowEngine class as only integration is needed to be tested."""

    def query(self, mf_request: MetricFlowQueryRequest) -> MetricFlowQueryResult:  # noqa: D
        data = {"metric1": [123, 5123, 23, 5123], "ds": [1, 2, 3, 4]}
        return MetricFlowQueryResult(
            query_spec=None,  # type: ignore
            dataflow_plan=None,  # type: ignore
            sql="SELECT * FROM hello",
            result_df=pd.DataFrame(data),
            result_table=None,
        )

    def dimensions_for_metrics(self, metric_names: List[str]) -> List[Dimension]:  # noqa: D
        dimensions = ["dim1", "dim2", "dim3"]
        return [Dimension(name) for name in dimensions]

    def list_metrics(self) -> List[Metric]:  # noqa: D
        return [Metric(name="metric1", dimensions=self.dimensions_for_metrics(metric_names=[]))]

    def get_dimension_values(  # noqa: D
        self,
        metric_name: str,
        get_group_by_values: str,
        time_constraint_start: Optional[datetime.datetime] = None,
        time_constraint_end: Optional[datetime.datetime] = None,
    ) -> List[str]:
        return ["dim_val1", "dim_val2", "dim_val3"]

    def list_materializations(self) -> List[Materialization]:  # noqa: D
        return [Materialization(name="mat1", metrics=["metric1"], dimensions=["dim1", "dim2"], destination_table=None)]


class MetricFlowCliRunner(CliRunner):
    """Custom CliRunner class to handle mocks."""

    def run(self, cli: click.BaseCommand, args: Sequence[str] = None) -> Result:  # noqa: D
        # Mock the metricflow engine
        cli_context = CLIContext()
        cli_context._mf = MockMetricFlowEngine()  # type: ignore
        return super().invoke(cli, args, obj=cli_context)


@pytest.fixture
def cli_runner() -> MetricFlowCliRunner:  # noqa: D
    return MetricFlowCliRunner()



import math
import pandas as pd


def assert_dataframes_equal(
    actual: pd.DataFrame, expected: pd.DataFrame, sort_columns: bool = True, allow_empty: bool = False
) -> None:
    """Check that contents of DataFrames are the same.

    If sort_columns is set to false, value and column order needs to be the same.
    """
    if set(actual.columns) != set(expected.columns):
        raise ValueError(
            f"DataFrames do not contain the same columns. actual: {set(actual.columns)}, "
            f"expected: {set(expected.columns)}"
        )

    if not allow_empty and actual.shape[0] == 0 and expected.shape[0] == 0:
        raise AssertionError("Both dataframes have no rows; likely there is a mistake with the test")

    equal = False
    # dataframes may be of different shapes causing equality check to fail
    if expected.shape == actual.shape:
        if sort_columns:
            sort_by = list(sorted(actual.columns.tolist()))
            expected = expected.loc[:, sort_by].sort_values(sort_by).reset_index(drop=True)
            actual = actual.loc[:, sort_by].sort_values(sort_by).reset_index(drop=True)
        equal = True
        # Compare all of the elements of the dataframe one by one. If someone knows pandas magic, this can be improved.
        for c in range(expected.shape[0]):
            for r in range(actual.shape[1]):
                # Floats need to be compared with an epsilon due to rounding error.
                # A % difference might be better.

                # NaNs can't be compared for equality.
                if pd.isna(expected.iloc[c, r]) and pd.isna(actual.iloc[c, r]):
                    equal = True
                elif isinstance(expected.iloc[c, r], float) and isinstance(actual.iloc[c, r], float):
                    equal = math.isclose(expected.iloc[c, r], actual.iloc[c, r])
                elif expected.iloc[c, r] != actual.iloc[c, r]:
                    equal = False

    if not equal:
        raise ValueError(
            f"Dataframes not equal.\n"
            f"Expected:\n{expected.to_markdown(index=False)}"
            "\n---\n"
            f"Actual:\n{actual.to_markdown(index=False)}"
        )



import re
from collections import OrderedDict
from typing import Dict, Tuple, List, Optional

from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.validations.validator_helpers import (
    ModelValidationRule,
    ValidationIssue,
    ValidationIssueLevel,
    validate_safely,
)
from metricflow.specs import ElementReference
from metricflow.time.time_granularity import TimeGranularity


class UniqueAndValidNameRule(ModelValidationRule):
    """Check that names are unique and valid.

    * Names of elements in data sources are unique / valid within the data source.
    * Names of data sources, dimension sets, metric sets, and materializations in the model are unique / valid.
    """

    NAME_REGEX = re.compile(r"\A[a-z][a-z0-9_]*[a-z0-9]\Z")

    @staticmethod
    def check_valid_name(name: str, data_source_name: Optional[str] = None) -> List[ValidationIssue]:  # noqa: D
        issues = []
        if not UniqueAndValidNameRule.NAME_REGEX.match(name):
            issues.append(
                ValidationIssue(
                    level=ValidationIssueLevel.ERROR,
                    model_object_reference=ValidationIssue.make_object_reference(
                        data_source_name=data_source_name,
                    )
                    if data_source_name
                    else OrderedDict(),
                    message=f"Invalid name `{name}` - names should only consist of lower case letters, numbers, "
                    f"and underscores. In addition, names should start with a lower case letter, and should not end "
                    f"with an underscore, and they must be at least 2 characters long.",
                )
            )
        if name.upper() in TimeGranularity.list_names():
            issues.append(
                ValidationIssue(
                    level=ValidationIssueLevel.ERROR,
                    model_object_reference=ValidationIssue.make_object_reference(
                        data_source_name=data_source_name,
                    )
                    if data_source_name
                    else OrderedDict(),
                    message=f"Invalid name `{name}` - names cannot match reserved time granularity keywords "
                    f"({TimeGranularity.list_names()})",
                )
            )
        return issues

    @staticmethod
    @validate_safely(whats_being_done="checking data source sub element names are unique")
    def _validate_data_source_elements(data_source: DataSource) -> List[ValidationIssue]:
        issues = []
        name_and_type_tuples: List[Tuple[ElementReference, str]] = []

        if data_source.measures:
            for measure in data_source.measures:
                name_and_type_tuples.append((measure.name, "measure"))
        if data_source.identifiers:
            for identifier in data_source.identifiers:
                name_and_type_tuples.append((identifier.name, "identifier"))
        if data_source.dimensions:
            for dimension in data_source.dimensions:
                name_and_type_tuples.append((dimension.name, "dimension"))

        name_to_type: Dict[ElementReference, str] = {}

        for name, _type in name_and_type_tuples:
            if name in name_to_type:
                issues.append(
                    ValidationIssue(
                        level=ValidationIssueLevel.FATAL,
                        model_object_reference=ValidationIssue.make_object_reference(
                            data_source_name=data_source.name,
                        ),
                        message=f"In data source `{data_source.name}`, can't use name `{name.element_name}` for a "
                        f"{_type} when it was already used for a {name_to_type[name]}",
                    )
                )
            else:
                name_to_type[name] = _type

        for name, _ in name_and_type_tuples:
            issues += UniqueAndValidNameRule.check_valid_name(data_source_name=data_source.name, name=name.element_name)

        return issues

    @staticmethod
    @validate_safely(whats_being_done="checking model top level element names are sufficiently unique")
    def _validate_top_level_objects(model: UserConfiguredModel) -> List[ValidationIssue]:
        """Checks names of objects that are not nested."""
        name_and_type_tuples = []
        if model.data_sources:
            for data_source in model.data_sources:
                name_and_type_tuples.append((data_source.name, "data source"))
        if model.materializations:
            for materialization in model.materializations:
                name_and_type_tuples.append((materialization.name, "materialization"))

        name_to_type: Dict[str, str] = {}

        issues = []

        for name, type_ in name_and_type_tuples:
            if name in name_to_type:
                issues.append(
                    ValidationIssue(
                        level=ValidationIssueLevel.FATAL,
                        model_object_reference=OrderedDict(),
                        message=f"Can't use name `{name}` for a {type_} when it was already used for a "
                        f"{name_to_type[name]}",
                    )
                )
            else:
                name_to_type[name] = type_

        if model.metrics:
            metric_names = set()
            for metric in model.metrics:
                if metric.name in metric_names:
                    issues.append(
                        ValidationIssue(
                            level=ValidationIssueLevel.FATAL,
                            model_object_reference=OrderedDict(),
                            message=f"Can't use name `{metric.name}` for a metric when it was already used for a metric",
                        )
                    )
                else:
                    metric_names.add(metric.name)

        if any([x.level == ValidationIssueLevel.FATAL for x in issues]):
            return issues

        for name, _ in name_and_type_tuples:
            issues += UniqueAndValidNameRule.check_valid_name(name)

        return issues

    @staticmethod
    @validate_safely(whats_being_done="running model validation ensuring elements have adequately unique names")
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssue]:  # noqa: D
        issues = []
        issues += UniqueAndValidNameRule._validate_top_level_objects(model=model)

        for data_source in model.data_sources:
            issues += UniqueAndValidNameRule._validate_data_source_elements(data_source=data_source)

        return issues



class SqlClientException(Exception):
    """Exception to represent errors related to the SqlClient"""

    def __init__(self) -> None:  # noqa: D
        super().__init__(
            "An error occurred when attempting to create/interact with the DB",
        )


class MetricFlowInitException(Exception):
    """Exception to represent errors related to the MetricFlow creation"""

    def __init__(self) -> None:  # noqa: D
        super().__init__(
            "An error occurred when attempting to create the MetricFlow engine",
        )


class ModelCreationException(Exception):
    """Exception to represent errors related to the building a model"""

    def __init__(self) -> None:  # noqa: D
        super().__init__(
            "An error occurred when attempting to build the semantic model",
        )



import logging
import os
from typing import TypeVar

import graphviz
from _pytest.fixtures import FixtureRequest

from metricflow.dag.mf_dag import DagNode, MetricFlowDag
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.plan_utils import snapshot_path_prefix

logger = logging.getLogger(__name__)


DagNodeT = TypeVar("DagNodeT", bound=DagNode)


def add_nodes_to_digraph(node: DagNodeT, dot: graphviz.Digraph) -> None:
    """Adds the node (and parent nodes) to the dot for visualization."""
    for parent_node in node.parent_nodes:
        add_nodes_to_digraph(parent_node, dot)

    dot.node(name=node.node_id.id_str, label=node.graphviz_label)
    for parent_node in node.parent_nodes:
        dot.edge(tail_name=parent_node.node_id.id_str, head_name=node.node_id.id_str)


DagGraphT = TypeVar("DagGraphT", bound=MetricFlowDag)


def display_graph_as_svg(
    mf_test_session_state: MetricFlowTestSessionState,
    request: FixtureRequest,
    dag_graph: DagGraphT,
) -> None:
    """Create and display the plan as an SVG, if configured to do so."""

    if not mf_test_session_state.display_plans:
        return
    plan_svg_output_path_prefix = snapshot_path_prefix(
        request=request, snapshot_group=dag_graph.__class__.__name__, snapshot_id=dag_graph.dag_id
    )

    dot = graphviz.Digraph(comment=dag_graph.dag_id, node_attr={"shape": "box", "fontname": "Courier"})
    dot.format = "svg"

    # Not quite correct if there are shared nodes.
    for sink_node in dag_graph.sink_nodes:
        add_nodes_to_digraph(sink_node, dot)

    # Create parent directory since it might not exist
    os.makedirs(os.path.dirname(plan_svg_output_path_prefix), exist_ok=True)

    dot.format = "svg"

    if mf_test_session_state.plans_displayed >= mf_test_session_state.max_plans_displayed:
        raise RuntimeError(
            f"Can't display plan - hit limit of " f"{mf_test_session_state.max_plans_displayed} plans displayed."
        )
    # Don't use the .svg extension since it will add it.
    dot.render(plan_svg_output_path_prefix, view=mf_test_session_state.display_plans, format="svg", cleanup=True)
    mf_test_session_state.plans_displayed += 1



import logging
import os
import pathlib
import yaml

from typing import Dict, Optional

from metricflow.protocols.sql_client import SqlClient
from metricflow.cli.constants import (
    CONFIG_DWH_URL,
    CONFIG_DWH_PASSWORD,
    CONFIG_DWH_SCHEMA,
    CONFIG_MODEL_PATH,
    CONFIG_PATH_KEY,
)
from metricflow.cli.exceptions import SqlClientException, MetricFlowInitException, ModelCreationException
from metricflow.engine.metricflow_engine import MetricFlowEngine
from metricflow.sql_clients.sql_utils import make_sql_client
from metricflow.cli.utils import ServerTimeSource
from metricflow.model.parsing.dir_to_model import parse_directory_of_yaml_files_to_model
from metricflow.model.semantic_model import SemanticModel
from metricflow.plan_conversion.column_resolver import DefaultColumnAssociationResolver
from metricflow.plan_conversion.time_spine import TimeSpineSource


# Avoid logs from being outputted to stdout
logging.getLogger("metricflow.sql_clients").setLevel(logging.CRITICAL)


class CLIContext:
    """Context for MetricFlow CLI."""

    def __init__(self) -> None:  # noqa: D
        self.verbose = False
        self._mf: Optional[MetricFlowEngine] = None
        self._sql_client: Optional[SqlClient] = None
        self._semantic_model: Optional[SemanticModel] = None
        self._mf_system_schema: Optional[str] = None
        self.config = ConfigHandler()

    @property
    def mf_system_schema(self) -> str:  # noqa: D
        if self._mf_system_schema is None:
            self._mf_system_schema = self.config._get_config_value(CONFIG_DWH_SCHEMA)
        assert self._mf_system_schema
        return self._mf_system_schema

    def __initialize_sql_client(self) -> None:
        """Initializes the SqlClient given the credentials."""
        try:
            url = self.config._get_config_value(CONFIG_DWH_URL)
            password = self.config._get_config_value(CONFIG_DWH_PASSWORD)
            self._sql_client = make_sql_client(url, password)
        except Exception as e:
            raise SqlClientException from e

    @property
    def sql_client(self) -> SqlClient:  # noqa: D
        if self._sql_client is None:
            # Initialize the SqlClient if not set
            self.__initialize_sql_client()
        assert self._sql_client is not None
        return self._sql_client

    def run_health_checks(self) -> Dict[str, Dict[str, str]]:
        """Execute the DB health checks."""
        try:
            return self.sql_client.health_checks(self.mf_system_schema)
        except Exception:
            raise SqlClientException

    def _initialize_metricflow_engine(self) -> None:
        """Initialize the MetricFlowEngine."""
        try:
            self._mf = MetricFlowEngine(
                semantic_model=self.semantic_model,
                sql_client=self.sql_client,
                column_association_resolver=DefaultColumnAssociationResolver(self.semantic_model),
                time_source=ServerTimeSource(),
                time_spine_source=TimeSpineSource(self.sql_client, schema_name=self.mf_system_schema),
                system_schema=self.mf_system_schema,
            )
        except Exception as e:
            raise MetricFlowInitException from e

    @property
    def mf(self) -> MetricFlowEngine:  # noqa: D
        if self._mf is None:
            self._initialize_metricflow_engine()
        assert self._mf is not None
        return self._mf

    def _build_semantic_model(self) -> None:
        """Get the path to the models and create a corresponding SemanticModel."""
        path_to_models = self.config._get_config_value(CONFIG_MODEL_PATH)
        try:
            model = parse_directory_of_yaml_files_to_model(path_to_models).model
            assert model is not None
        except Exception as e:
            raise ModelCreationException from e
        self._semantic_model = SemanticModel(model)

    @property
    def semantic_model(self) -> SemanticModel:  # noqa: D
        if self._semantic_model is None:
            self._build_semantic_model()
        assert self._semantic_model is not None
        return self._semantic_model


class ConfigHandler:
    """Class to handle all config file retrieval/insertion actions."""

    def __init__(self) -> None:  # noqa: D
        # Create config directory if not exist
        if not os.path.exists(self.dir_path):
            dir_path = pathlib.Path(self.dir_path)
            dir_path.mkdir(parents=True)

    @property
    def dir_path(self) -> str:
        """Retrieve MetricFlow config directory from $MF_CONFIG_DIR, default config dir is ~/.metricflow"""
        config_dir_env = os.getenv(CONFIG_PATH_KEY)
        return config_dir_env if config_dir_env else f"{str(pathlib.Path.home())}/.metricflow"

    @property
    def file_path(self) -> str:
        """Config file can be found at <config_dir>/config.yml"""
        return os.path.join(self.dir_path, "config.yml")

    def _get_config_value(self, key: str) -> str:
        """Attempt to get a corresponding value from the config file. Throw an error if not exists or None."""
        config_file = open(self.file_path, "r")

        config = yaml.load(config_file, Loader=yaml.Loader)
        if key not in config:
            raise KeyError(f"{key} is missing in the configuration file.")

        value = config[key]
        if value is None:
            raise ValueError(f"value for {key} cannot be None.")
        return value



from enum import Enum
from typing import Tuple
from dataclasses import dataclass
from metricflow.model.objects.elements.identifier import IdentifierType
from metricflow.model.objects.common import Element


class JoinType(Enum):  # noqa: D
    INNER = "inner"
    LEFT = "left"


@dataclass(frozen=True)
class JoinLink:  # noqa: D
    via_from: Element  # identifier for join (left side "from")
    via_to: Element  # identifier for join (right side "target")
    partitions: Tuple[str, ...]  # partitions for join, tuple for hashability
    join_type: JoinType

    @property
    def is_fanout_join(self) -> bool:  # noqa: D
        return (self.via_from.type, self.via_to.type) in FANOUT_JOIN_TYPES  # type: ignore


FANOUT_JOIN_TYPES = set(
    [
        (IdentifierType.PRIMARY, IdentifierType.FOREIGN),
        (IdentifierType.UNIQUE, IdentifierType.FOREIGN),
        (IdentifierType.FOREIGN, IdentifierType.FOREIGN),
    ]
)

JOIN_TYPE_MAPPING = {
    (IdentifierType.PRIMARY, IdentifierType.PRIMARY): JoinType.LEFT,
    (IdentifierType.PRIMARY, IdentifierType.UNIQUE): JoinType.LEFT,
    # (IdentifierType.PRIMARY, IdentifierType.FOREIGN): JoinType.LEFT,  # fans out
    (IdentifierType.UNIQUE, IdentifierType.PRIMARY): JoinType.LEFT,
    (IdentifierType.UNIQUE, IdentifierType.UNIQUE): JoinType.LEFT,
    # (IdentifierType.UNIQUE, IdentifierType.FOREIGN): JoinType.LEFT,  # fans out
    (IdentifierType.FOREIGN, IdentifierType.PRIMARY): JoinType.LEFT,
    (IdentifierType.FOREIGN, IdentifierType.UNIQUE): JoinType.LEFT,
    # (IdentifierType.FOREIGN, IdentifierType.FOREIGN): JoinType.LEFT,  # fans out
}



from __future__ import annotations

import json
from abc import ABC, abstractmethod

from pydantic import BaseModel


class HashableBaseModel(BaseModel):
    """Extends BaseModel with a generic hash function"""

    def __hash__(self) -> int:  # noqa: D
        return hash(json.dumps(self.json(sort_keys=True), sort_keys=True))


class FrozenBaseModel(HashableBaseModel):
    """Similar to HashableBaseModel but faux immutable."""

    class Config:
        """Pydantic feature."""

        allow_mutation = False

    def to_pretty_json(self) -> str:
        """Convert to a pretty JSON representation."""
        raw_json_str = self.json()
        json_obj = json.loads(raw_json_str)
        return json.dumps(json_obj, indent=4)

    def __str__(self) -> str:  # noqa: D
        return self.__repr__()


class ParseableObject:  # noqa: D
    pass


class ParseableField(ABC):  # noqa: D
    @staticmethod
    @abstractmethod
    def parse(s: str) -> ParseableField:  # noqa: D
        pass



import pytest

from metricflow.model.objects.constraints.where import WhereClauseConstraint


@pytest.mark.skip(reason="Does not currently parse correctly.")
def test_where_constraint_parsing() -> None:  # noqa: D
    """Currently throws an exception:

    ConstraintParseException: expected parsed constraint to contain exactly one key; got {}
    """
    parsed_where = WhereClauseConstraint.parse("ds = CAST('2020-01-01' AS TIMESTAMP) AND is_instant")
    assert set(parsed_where.linkable_names) == {"ds", "is_instant"}



from metricflow.test.fixtures.cli_fixtures import MetricFlowCliRunner
from metricflow.cli.main import (
    get_dimension_values,
    list_dimensions,
    list_materializations,
    list_metrics,
    query,
)


def test_query(cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    resp = cli_runner.run(query, args=["--metrics", "metric1", "--dimensions", "ds"])

    assert "metric1" in resp.output
    assert resp.exit_code == 0


def test_list_dimensions(cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    resp = cli_runner.run(list_dimensions, args=["--metric-names", "metric1"])

    assert "dim1" in resp.output
    assert resp.exit_code == 0


def test_list_metrics(cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    resp = cli_runner.run(list_metrics)

    assert "metric1: dim1, dim2, dim3" in resp.output
    assert resp.exit_code == 0


def test_get_dimension_values(cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    resp = cli_runner.run(get_dimension_values, args=["--metric-name", "metric1", "--dimension-name", "ds"])

    assert "dim_val1" in resp.output
    assert resp.exit_code == 0


def test_list_materializations(cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    resp = cli_runner.run(list_materializations)

    assert "materialization: details related to materialization" in resp.output
    assert resp.exit_code == 0



import click
import datetime as dt
import traceback

from dateutil.parser import parse
from functools import wraps
from typing import Any, Callable, List, Optional

from metricflow.cli.constants import CONFIG_DWH_URL, CONFIG_DWH_PASSWORD, CONFIG_DWH_SCHEMA, CONFIG_MODEL_PATH
from metricflow.time.time_source import TimeSource

CONFIG_TEMPLATE = {
    CONFIG_DWH_URL: "",
    CONFIG_DWH_PASSWORD: "",
    CONFIG_DWH_SCHEMA: "",
    CONFIG_MODEL_PATH: "",
}


# Click Options
def query_options(function: Callable) -> Callable:
    """Common options for a query"""
    function = click.option(
        "--order",
        required=False,
        multiple=True,
        help='Metrics or dimensions to order by ("-" in front of a column means descending). For example: --order -ds',
    )(function)

    function = click.option(
        "--limit",
        required=False,
        type=str,
        help="Limit the number of rows out using an int or leave blank for no limit. For example: --limit 100",
        callback=lambda ctx, param, value: validate_limit(value),
    )(function)

    function = click.option(
        "--where",
        required=False,
        type=str,
        default=None,
        help='SQL-like where statement provided as a string. For example: --where "revenue > 100"',
    )(function)

    function = start_end_time_options(function)
    return function


def start_end_time_options(function: Callable) -> Callable:
    """Options for start_time and end_time."""
    function = click.option(
        "--start-time",
        required=False,
        type=str,
        default=None,
        help="Optional iso8601 timestamp to constraint the start time of the data (inclusive)",
        callback=lambda ctx, param, value: convert_to_datetime(value),
    )(function)

    function = click.option(
        "--end-time",
        required=False,
        type=str,
        default=None,
        help="Optional iso8601 timestamp to constraint the end time of the data (inclusive)",
        callback=lambda ctx, param, value: convert_to_datetime(value),
    )(function)
    return function


def separated_by_comma_option(option_name: str, help_msg: str = "") -> Callable:
    """Parse input containing a string separated by commma to a List."""

    def wraps(function: Callable) -> Callable:
        function = click.option(
            option_name,
            required=True,
            help=help_msg,
            callback=lambda ctx, param, value: parse_comma_separated_inputs(value),
        )(function)
        return function

    return wraps


# Parsers/Validators
def convert_to_datetime(datetime_str: Optional[str]) -> Optional[dt.datetime]:
    """Callback to convert string to datetime given as an iso8601 timestamp."""
    if datetime_str is None:
        return None

    if not valid_datetime(datetime_str):
        raise click.BadParameter("must be valid iso8601 timestamp")

    return dt.datetime.fromisoformat(datetime_str)


def valid_datetime(dt_str: str) -> bool:
    """Returns true if string is valid iso8601 timestamp, false otherwise"""
    try:
        parse(dt_str)
    except Exception:
        return False
    return True


def parse_comma_separated_inputs(value: str) -> List[str]:  # noqa: D
    # If comma exist, explode this into a list and return
    if "," in value:
        return [i.strip() for i in value.split(",")]

    # Return a list of the single value
    return [value]


def validate_limit(limit: Optional[str]) -> Optional[int]:
    """Validates and transform limit input."""
    if limit and not limit.isnumeric():
        raise click.BadParameter("limit must be an int. For no limit, do not pass this argument")
    return int(limit) if limit else None


# Misc
def exception_handler(func: Callable[..., Any]) -> Callable[..., Any]:  # type: ignore[misc]
    """Decorator to handle exceptions."""

    @wraps(func)
    def wrapper(*args, **kwargs):  # type: ignore
        try:
            func(*args, **kwargs)
        except Exception as e:
            # Checks if CLIContext has verbose flag set
            if args and hasattr(args[0], "verbose") and args[0].verbose is True:
                click.echo(traceback.format_exc())
            else:
                click.echo(f"\nERROR: {str(e)}")
            exit(1)

    return wrapper


class ServerTimeSource(TimeSource):
    """A time source that represents the current datetime in UTC."""

    def get_time(self) -> dt.datetime:  # noqa: D
        return dt.datetime.utcnow()



import click
import datetime as dt
import jinja2
import pandas as pd
import pathlib
import textwrap
import time
import yaml

from halo import Halo
from packaging.version import parse
from typing import List, Optional
from update_checker import UpdateChecker

from metricflow.cli import PACKAGE_NAME, __version__
from metricflow.cli.cli_context import CLIContext
from metricflow.cli.utils import (
    CONFIG_TEMPLATE,
    exception_handler,
    query_options,
    separated_by_comma_option,
    start_end_time_options,
)
from metricflow.dataflow.dataflow_plan_to_text import dataflow_plan_as_text
from metricflow.engine.metricflow_engine import MetricFlowQueryRequest


MAX_LIST_OBJECT_ELEMENTS = 5
pass_config = click.make_pass_decorator(CLIContext, ensure=True)


@click.group()
@click.option("-v", "--verbose", is_flag=True)
@pass_config
def cli(cfg: CLIContext, verbose: bool) -> None:  # noqa: D
    cfg.verbose = verbose

    checker = UpdateChecker()
    result = checker.check(PACKAGE_NAME, __version__)
    # result is None when an update was not found or a failure occured
    if result:
        click.secho(
            "Warning: A new version of the MetricFlow CLI is available.",
            bold=True,
            fg="red",
        )

        click.echo(
            f"Please update to version {result.available_version}, released {result.release_date} by running:\n\t$ pip install --upgrade {PACKAGE_NAME}",
        )


@cli.command()
def version() -> None:
    """Print the current version of the MetricFlow CLI."""
    click.echo(__version__)


@cli.command()
@pass_config
def setup(cfg: CLIContext) -> None:
    """Setup MetricFlow."""

    click.echo(
        textwrap.dedent(
            """\
            Welcome to MetricFlow!
            """
        )
    )

    path = pathlib.Path(cfg.config.file_path)
    abs_path = path.absolute()
    to_create = not path.exists()

    # Seed the config template to the config file
    if to_create:
        with open(abs_path, "w") as file:
            file.write(yaml.dump(CONFIG_TEMPLATE))

    click.echo(
        textwrap.dedent(
            f"""\
            A template config file has {'' if to_create else 'already '}been created in {abs_path}.

              1. Fill it out with the relevant details
              2. run `mf health-check` to validate the Data Warehouse connection
              3. run `mf validate-config` to validate the Model configurations

              or run `mf validate-setup` to run all necessary checks
            """
        )
    )


@cli.command()
@separated_by_comma_option(
    "--metrics",
    "Metrics to query for: syntax is --metrics bookings or for multiple metrics --metrics bookings,messages",
)
@separated_by_comma_option(
    "--dimensions", "Dimensions to group by: syntax is --dimensions ds or for multiple dimensions --dimensions ds,org"
)
@query_options
@click.option(
    "--as-table",
    required=False,
    type=str,
    help="Output the data to a specified SQL table in the form of '<schema>.<table>'",
)
@click.option(
    "--csv",
    type=click.File("wb"),
    required=False,
    help="Provide filepath for dataframe output to csv",
)
@click.option(
    "--explain",
    is_flag=True,
    required=False,
    default=False,
    help="In the query output, show the query that was executed against the data warehouse",
)
@click.option(
    "--decimals",
    required=False,
    default=2,
    help="Choose the number of decimal places to round for the numerical values",
)
@pass_config
@exception_handler
def query(
    cfg: CLIContext,
    metrics: List[str],
    dimensions: List[str],
    where: Optional[str] = None,
    start_time: Optional[dt.datetime] = None,
    end_time: Optional[dt.datetime] = None,
    order: Optional[List[str]] = None,
    limit: Optional[int] = None,
    as_table: Optional[str] = None,
    csv: Optional[click.utils.LazyFile] = None,
    explain: bool = False,
    decimals: int = 2,
) -> None:
    """Create a new query with MetricFlow and assembles a MetricFlowQueryResult."""
    start = time.time()
    spinner = Halo(text="Initiating query…", spinner="dots")
    spinner.start()

    mf_request = MetricFlowQueryRequest.create_with_random_request_id(
        metric_names=metrics,
        group_by_names=dimensions,
        limit=limit,
        time_constraint_start=start_time,
        time_constraint_end=end_time,
        where_constraint=where,
        order_by_names=order,
        output_table=as_table,
    )

    if explain:
        explain_result = cfg.mf.explain(mf_request=mf_request)
    else:
        query_result = cfg.mf.query(mf_request=mf_request)

    spinner.succeed(f"Success 🦄 - query completed after {time.time() - start:.2f} seconds")

    if explain:
        sql = explain_result.rendered_sql.sql_query
        click.echo("🔎 Generated Dataflow Plan + SQL (remove --explain to see data):")
        click.echo(
            textwrap.indent(
                jinja2.Template(
                    textwrap.dedent(
                        """\
                        Dataflow Plan:
                            {{ plan_text | indent(4) }}
                        """
                    ),
                    undefined=jinja2.StrictUndefined,
                ).render(plan_text=dataflow_plan_as_text(explain_result.dataflow_plan)),
                prefix="-- ",
            )
        )
        click.echo(sql)
        exit()

    df = query_result.result_df
    # Show the data if returned successfully
    if df is not None:
        if df.empty:
            click.echo("Successful MQL query returned an empty result set.")
        elif csv is not None:
            df.to_csv(csv, index=False)
            click.echo(f"Successfully written query output to {csv.name}")
        else:
            # NOTE: remove `to_string` if no pandas dependency is < 1.1.0
            if parse(pd.__version__) >= parse("1.1.0"):
                click.echo(df.to_markdown(index=False, floatfmt=f".{decimals}f"))
            else:
                click.echo(df.to_string(index=False, float_format=lambda x: format(x, f".{decimals}f")))


@cli.command()
@click.option("--search", required=False, type=str, help="Filter available metrics by this search term")
@click.option("--show-all-dims", is_flag=True, default=False, help="Show all dimensions associated with a metric.")
@pass_config
@exception_handler
def list_metrics(cfg: CLIContext, show_all_dims: bool = False, search: Optional[str] = None) -> None:
    """List the metrics with their available dimensions.

    Automatically truncates long lists of dimensions, pass --show-all-dims to see all.
    """

    spinner = Halo(text="Looking for all available metrics...", spinner="dots")
    spinner.start()

    metrics = cfg.mf.list_metrics()

    if not metrics:
        spinner.fail("List of metrics unavailable.")

    filter_msg = ""
    if search is not None:
        num_metrics = len(metrics)
        metrics = [m for m in metrics if search.lower() in m.name.lower()]
        filter_msg = f" matching `{search}`, of a total of {num_metrics} available"

    spinner.succeed(f"🌱 We've found {len(metrics)} metrics{filter_msg}.")
    click.echo('The list below shows metrics in the format of "metric_name: list of available dimensions"')
    num_dims_to_show = MAX_LIST_OBJECT_ELEMENTS
    for m in metrics:
        # sort dimensions by whether they're local first(if / then global else local) then the dim name
        dimensions = sorted(map(lambda d: d.name, filter(lambda d: "/" not in d.name, m.dimensions))) + sorted(
            map(lambda d: d.name, filter(lambda d: "/" in d.name, m.dimensions))
        )
        if show_all_dims:
            num_dims_to_show = len(dimensions)
        click.echo(
            f"• {click.style(m.name, bold=True, fg='green')}: {', '.join(dimensions[:num_dims_to_show])}"
            + (f" and {len(dimensions) - num_dims_to_show} more" if len(dimensions) > num_dims_to_show else "")
        )


@cli.command()
@separated_by_comma_option(
    "--metric-names", help_msg="List dimensions by given metrics (intersection). Ex. --metric-names bookings,messages"
)
@pass_config
@exception_handler
def list_dimensions(cfg: CLIContext, metric_names: List[str]) -> None:
    """List all unique dimensions."""
    spinner = Halo(
        text="Looking for all available dimensions...",
        spinner="dots",
    )
    spinner.start()

    dimensions = cfg.mf.dimensions_for_metrics(metric_names)
    if not dimensions:
        spinner.fail("List of dimensions unavailable.")

    spinner.succeed(f"🌱 We've found {len(dimensions)} common dimensions for metrics {metric_names}.")
    for d in dimensions:
        click.echo(f"• {click.style(d.name, bold=True, fg='green')}")


@cli.command()
@pass_config
@exception_handler
def health_checks(cfg: CLIContext) -> None:
    """Performs a health check against the DW provided in the configs."""
    spinner = Halo(
        text="Running health checks against your data warehouse... (This should not take longer than 30s for a successful connection)",
        spinner="dots",
    )
    spinner.start()
    res = cfg.run_health_checks()
    spinner.succeed("Health checks completed.")
    for test in res:
        test_res = res[test]
        if test_res["status"] != "SUCCESS":
            click.echo(f"• ❌ {click.style(test, bold=True, fg=('red'))}:  Failed with - {test_res['error_message']}.")
        else:
            click.echo(f"• ✅ {click.style(test, bold=True, fg=('green'))}: Success!")


@cli.command()
@click.option("--dimension-name", required=True, type=str, help="Dimension to query values from")
@click.option("--metric-name", required=True, type=str, help="Metric that is associated with the dimension")
@start_end_time_options
@pass_config
@exception_handler
def get_dimension_values(
    cfg: CLIContext,
    metric_name: str,
    dimension_name: str,
    start_time: Optional[dt.datetime] = None,
    end_time: Optional[dt.datetime] = None,
) -> None:
    """List all dimension values with the corresponding metric."""
    spinner = Halo(
        text=f"Retrieving dimension values for dimension '{dimension_name}' of metric '{metric_name}'...",
        spinner="dots",
    )
    spinner.start()

    try:
        dim_vals = cfg.mf.get_dimension_values(
            metric_name=metric_name,
            get_group_by_values=dimension_name,
            time_constraint_start=start_time,
            time_constraint_end=end_time,
        )
    except Exception as e:
        spinner.fail()
        click.echo(
            textwrap.dedent(
                f"""\
                ❌ Failed to query dimension values for dimension {dimension_name} of metric {metric_name}.
                    ERROR: {str(e)}
                """
            )
        )
        exit(1)

    spinner.succeed(
        f"🌱 We've found {len(dim_vals)} dimension values for dimension {dimension_name} of metric {metric_name}."
    )
    for dim_val in dim_vals:
        click.echo(f"• {click.style(dim_val, bold=True, fg='green')}")


@cli.command()
@click.option("--search", required=False, type=str, help="Filter available materializations by this search term")
@pass_config
@exception_handler
def list_materializations(cfg: CLIContext, search: Optional[str] = None) -> None:
    """List the materializations with their available metrics and dimensions."""

    spinner = Halo(text="Looking for all available materializations...", spinner="dots")
    spinner.start()

    materializations = cfg.mf.list_materializations()
    if not materializations:
        spinner.fail("List of materializations unavailable.")

    filter_msg = ""
    if search is not None:
        count = len(materializations)
        materializations = [m for m in materializations if search.lower() in m.name.lower()]
        filter_msg = f" matching `{search}`, of a total of {count} available"

    spinner.succeed(f"🌱 We've found {len(materializations)} materializations{filter_msg}.")
    click.echo(
        'The list below shows materializations in the format of "materialization: details related to materialization"'
    )
    for m in materializations:
        dimensions = sorted(m.dimensions)
        metrics = sorted(m.metrics)
        # Materialization name
        click.echo(f"• {click.style(m.name, bold=True, fg='green')}:")
        # Metrics related to this materalization
        click.echo(
            f"Metrics: {', '.join(metrics[:MAX_LIST_OBJECT_ELEMENTS])}"
            + (
                f" and {len(metrics) - MAX_LIST_OBJECT_ELEMENTS} more"
                if len(metrics) > MAX_LIST_OBJECT_ELEMENTS
                else ""
            )
        )
        # Dimensions related to this materalization
        click.echo(
            f"Dimensions: {', '.join(dimensions[:MAX_LIST_OBJECT_ELEMENTS])}"
            + (
                f" and {len(dimensions) - MAX_LIST_OBJECT_ELEMENTS} more"
                if len(dimensions) > MAX_LIST_OBJECT_ELEMENTS
                else ""
            )
        )
        # Misc related to this materialization
        click.echo(f"destination table: {m.destination_table or m.name}")


@cli.command()
@click.option(
    "--materialization-name",
    required=True,
    type=str,
    help="Name of materialization to materialize",
)
@start_end_time_options
@pass_config
@exception_handler
def materialize(
    cfg: CLIContext,
    materialization_name: str,
    start_time: Optional[dt.datetime] = None,
    end_time: Optional[dt.datetime] = None,
) -> None:
    """Create a new materialization query and returns materialized table"""
    if start_time is None and not click.confirm(
        "You haven't provided a start_time. This means we will materialize from the beginning of time. This may be expensive. Are you sure you want to continue?"
    ):
        click.echo("Exiting")
        return

    start = time.time()
    spinner = Halo(text="Initiating materialization query…", spinner="dots")
    spinner.start()

    result_table = cfg.mf.materialize(
        materialization_name,
        time_constraint_start=start_time,
        time_constraint_end=end_time,
    )

    spinner.succeed(f"Success 🦄 - materialize query completed after {time.time() - start:.2f} seconds.")
    click.echo(f"Materialized table created at: {result_table.sql}")


@cli.command()
@click.option(
    "--materialization-name",
    required=True,
    type=str,
    help="Name of materialization to drop",
)
@pass_config
@exception_handler
def drop_materialization(cfg: CLIContext, materialization_name: str) -> None:
    """Drops a given materialized table."""

    start = time.time()
    spinner = Halo(text="Initiating drop materialization query…", spinner="dots")
    spinner.start()

    result = cfg.mf.drop_materialization(materialization_name=materialization_name)

    if result:
        spinner.succeed(f"Success 🦄 - drop materialization query completed after {time.time() - start:.2f} seconds.")
    else:
        spinner.warn(f"Materialized table for `{materialization_name}` did not exist, no table was dropped")


if __name__ == "__main__":
    cli()



import logging

from metricflow.model.semantic_model import SemanticModel
from metricflow.model.semantics.linkable_spec_resolver import (
    ValidLinkableSpecResolver,
    LinkableElementProperties,
)
from metricflow.model.semantics.semantic_containers import MAX_JOIN_HOPS
from metricflow.specs import MetricSpec

logger = logging.getLogger(__name__)


def test_linkable_spec_resolver(simple_semantic_model: SemanticModel) -> None:  # noqa: D
    resolver = ValidLinkableSpecResolver(
        user_configured_model=simple_semantic_model.user_configured_model,
        primary_time_dimension_reference=simple_semantic_model.data_source_semantics.primary_time_dimension_reference,
        max_identifier_links=MAX_JOIN_HOPS,
    )

    result = resolver.get_linkable_elements_for_metrics(
        metric_specs=[MetricSpec(element_name="bookings"), MetricSpec(element_name="views")],
        with_any_of=LinkableElementProperties.all_properties(),
        without_any_of=frozenset({LinkableElementProperties.DERIVED_TIME_GRANULARITY}),
    ).as_spec_set

    assert [
        "create_a_cycle_in_the_join_graph__is_instant",
        "create_a_cycle_in_the_join_graph__listing__capacity_latest",
        "create_a_cycle_in_the_join_graph__listing__country_latest",
        "create_a_cycle_in_the_join_graph__listing__is_lux_latest",
        "listing__capacity_latest",
        "listing__country_latest",
        "listing__is_lux_latest",
        "listing__user__company_name",
        "listing__user__home_state",
        "listing__user__home_state_latest",
    ] == sorted(tuple(x.qualified_name for x in result.dimension_specs))
    assert [
        "create_a_cycle_in_the_join_graph__listing__created_at",
        "create_a_cycle_in_the_join_graph__listing__ds",
        "ds",
        "ds_partitioned",
        "listing__created_at",
        "listing__ds",
        "listing__user__created_at",
        "listing__user__ds_partitioned",
    ] == sorted(tuple(x.qualified_name for x in result.time_dimension_specs))
    assert [
        "create_a_cycle_in_the_join_graph",
        "create_a_cycle_in_the_join_graph__listing",
        "create_a_cycle_in_the_join_graph__listing__lux_listing",
        "create_a_cycle_in_the_join_graph__listing__user",
        "listing",
        "listing__lux_listing",
        "listing__user",
        "listing__user__company",
    ] == sorted(tuple(x.qualified_name for x in result.identifier_specs))



from __future__ import annotations

import logging
from collections import defaultdict
from copy import deepcopy
from typing import Dict, List, Set, Type, Optional, Sequence, Tuple

from metricflow.errors.errors import (
    DuplicateMetricError,
    MetricNotFoundError,
    NonExistentMeasureError,
    InvalidDataSourceError,
)
from metricflow.instances import DataSourceReference, DataSourceElementReference
from metricflow.model.graph import MultiDiGraph
from metricflow.model.objects.common import Element
from metricflow.model.objects.data_source import DataSource, DataSourceOrigin
from metricflow.model.objects.elements.dimension import Dimension, DimensionType
from metricflow.model.objects.elements.measure import Measure, AggregationType
from metricflow.model.objects.metric import Metric, MetricType
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.semantics.data_source_container import PydanticDataSourceContainer
from metricflow.model.semantics.linkable_spec_resolver import (
    ValidLinkableSpecResolver,
    LinkableElementProperties,
)
from metricflow.model.semantics.links_new import JoinLink, JOIN_TYPE_MAPPING
from metricflow.specs import (
    LinkableInstanceSpec,
    LinkableElementReference,
    MeasureSpec,
    MeasureReference,
    DimensionReference,
    IdentifierReference,
    MetricSpec,
    TimeDimensionReference,
    LinkableSpecSet,
)

logger = logging.getLogger(__name__)

MAX_JOIN_HOPS = 2


class MetricSemantics:  # noqa: D
    def __init__(  # noqa: D
        self, user_configured_model: UserConfiguredModel, data_source_semantics: DataSourceSemantics
    ) -> None:
        self._user_configured_model = user_configured_model
        self._metrics: Dict[MetricSpec, Metric] = {}
        self._data_source_semantics = data_source_semantics

        # Dict from the name of the metric to the hash.
        self._metric_hashes: Dict[MetricSpec, str] = {}

        for metric in self._user_configured_model.metrics:
            self.add_metric(metric)

        self._linkable_spec_resolver = ValidLinkableSpecResolver(
            user_configured_model=self._user_configured_model,
            primary_time_dimension_reference=self._data_source_semantics.primary_time_dimension_reference,
            max_identifier_links=MAX_JOIN_HOPS,
        )

    def element_specs_for_metrics(
        self,
        metric_specs: List[MetricSpec],
        local_only: bool = False,
        dimensions_only: bool = False,
        exclude_multi_hop: bool = False,
        exclude_derived_time_granularities: bool = True,
        exclude_local_linked_primary_time: bool = True,
    ) -> List[LinkableInstanceSpec]:
        """Dimensions common to all metrics requested (intersection)"""
        without_any_property = set()

        if exclude_local_linked_primary_time:
            without_any_property.add(LinkableElementProperties.LOCAL_LINKED_PRIMARY_TIME)

        if exclude_derived_time_granularities:
            without_any_property.add(LinkableElementProperties.DERIVED_TIME_GRANULARITY)

        if local_only:
            with_any_property = frozenset({LinkableElementProperties.LOCAL})
        else:
            with_any_property = LinkableElementProperties.all_properties()

        if exclude_multi_hop:
            without_any_property.add(LinkableElementProperties.MULTI_HOP)

        all_linkable_specs = self._linkable_spec_resolver.get_linkable_elements_for_metrics(
            metric_specs=metric_specs,
            with_any_of=with_any_property,
            without_any_of=frozenset(without_any_property),
        ).as_spec_set

        if dimensions_only:
            all_linkable_specs = LinkableSpecSet(
                dimension_specs=all_linkable_specs.dimension_specs,
                time_dimension_specs=all_linkable_specs.time_dimension_specs,
            )

        return sorted(all_linkable_specs.as_tuple, key=lambda x: x.qualified_name)

    def get_metrics(self, metric_names: List[MetricSpec]) -> List[Metric]:  # noqa: D
        res = []
        for metric_name in metric_names:
            if metric_name not in self._metrics:
                raise MetricNotFoundError(f"Unable to find metric `{metric_name}`. Perhaps it has not been registered")
            res.append(self._metrics[metric_name])

        return res

    @property
    def metric_names(self) -> List[MetricSpec]:  # noqa: D
        return list(self._metrics.keys())

    def get_metric(self, metric_name: MetricSpec) -> Metric:  # noqa:D
        if metric_name not in self._metrics:
            raise MetricNotFoundError(f"Unable to find metric `{metric_name}`. Perhaps it has not been registered")
        return self._metrics[metric_name]

    def add_metric(self, metric: Metric) -> None:
        """Add metric, validating presence of required measures"""
        metric_spec = MetricSpec(element_name=metric.name)
        if metric_spec in self._metrics:
            raise DuplicateMetricError(f"Metric `{metric.name}` has already been registered")
        for measure_reference in metric.measure_names:
            if measure_reference not in self._data_source_semantics.measure_references:
                raise NonExistentMeasureError(
                    f"Metric `{metric.name}` references measure `{measure_reference}` which has not been registered"
                )
        self._metrics[metric_spec] = metric
        self._metric_hashes[metric_spec] = metric.definition_hash

    @property
    def valid_hashes(self) -> Set[str]:
        """Return all of the hashes of the metric definitions."""
        return set(self._metric_hashes.values())

    def measures_for_metric(self, metric_spec: MetricSpec) -> Tuple[MeasureSpec, ...]:
        """Return the measure specs required to compute the metric."""
        metric = self.get_metric(metric_spec)

        return tuple(
            MeasureSpec(
                element_name=x.element_name,
            )
            for x in metric.measure_names
        )

    def contains_cumulative_metric(self, metric_specs: Sequence[MetricSpec]) -> bool:
        """Returns true if any of the specs correspond to a cumulative metric."""
        for metric_spec in metric_specs:
            if self.get_metric(metric_spec).type == MetricType.CUMULATIVE:
                return True
        return False


class DataSourceSemantics:
    """Tracks semantic information for data source held in a set of DataSourceContainers"""

    def __init__(  # noqa: D
        self,
        model: UserConfiguredModel,
        configured_data_source_container: PydanticDataSourceContainer,
    ) -> None:
        self._model = model
        self._element_types: Dict[LinkableElementReference, Type[Element]] = {}
        self._measure_index: Dict[MeasureReference, List[DataSource]] = defaultdict(list)
        self._measure_aggs: Dict[
            MeasureReference, AggregationType
        ] = {}  # maps measures to their one consistent aggregation
        self._dimension_index: Dict[DimensionReference, List[DataSource]] = defaultdict(list)
        self._linkable_reference_index: Dict[LinkableElementReference, List[DataSource]] = defaultdict(list)
        self._entity_index: Dict[Optional[str], List[DataSource]] = defaultdict(list)
        self._identifier_ref_to_entity: Dict[IdentifierReference, Optional[str]] = {}
        # Graph where nodes are DataSource names and edges are Join objects that represent all valid
        # joins possible between the nodes (data sources)
        self._data_source_links = MultiDiGraph[str, JoinLink]()

        self._data_source_names: Set[str] = set()

        self._configured_data_source_container = configured_data_source_container

        # Add semantic tracking for data sources from configured_data_source_container
        for data_source in self._configured_data_source_container.values():
            assert isinstance(data_source, DataSource)
            self.add_configured_data_source(data_source)

    @property
    def data_sources(self) -> Sequence[DataSource]:  # noqa: D
        return self._model.data_sources

    def add_configured_data_source(self, data_source: DataSource) -> None:
        """Dont use this unless you mean it (ie in tests). The configured data sources are supposed to be static"""
        self._configured_data_source_container._put(data_source)
        self.add_data_source(data_source)

    def get_linkable_element_references(self) -> List[LinkableElementReference]:  # noqa: D
        return list(self._linkable_reference_index.keys())

    def get_dimension_references(self) -> List[DimensionReference]:  # noqa: D
        return list(self._dimension_index.keys())

    def get_linkable(  # noqa: D
        self,
        linkable_name: LinkableElementReference,
        origin: Optional[DataSourceOrigin] = None,
        use_identifier_dundered_name: bool = False,
    ) -> Element:
        """Retrieves a full linkable element object by name

        use_identifier_dundered_name returns the Dimension with the identifier-dundered name (if the requested linkable element
        has an identifier-dunder) otherwise the identifier is stripped from the name
        """
        for dimension_source in self._linkable_reference_index[linkable_name]:
            if origin and dimension_source.origin != origin:
                continue
            linkable_element = dimension_source.get_element(linkable_name)
            # find the data source that has the requested linkable_element by the requested identifier

            return deepcopy(linkable_element)

        raise ValueError(
            f"Could not find dimension with name ({linkable_name.element_name}) in configured data sources"
        )

    def get_dimension(
        self, dimension_reference: DimensionReference, origin: Optional[DataSourceOrigin] = None
    ) -> Dimension:
        """Retrieves a full dimension object by name"""
        for dimension_source in self._dimension_index[dimension_reference]:
            if origin and dimension_source.origin != origin:
                continue
            dimension = dimension_source.get_dimension(dimension_reference)
            # find the data source that has the requested dimension by the requested identifier

            return deepcopy(dimension)

        raise ValueError(
            f"Could not find dimension with name ({dimension_reference.element_name}) in configured data sources"
        )

    @property
    def measure_references(self) -> List[MeasureReference]:  # noqa: D
        return list(self._measure_index.keys())

    def get_measure(self, measure_reference: MeasureReference) -> Measure:  # noqa: D
        if measure_reference not in self._measure_index:
            raise ValueError(f"Could not find measure with name ({measure_reference}) in configured data sources")

        assert len(self._measure_index[measure_reference]) >= 1
        # Measures should be consistent across data sources, so just use the first one.
        return list(self._measure_index[measure_reference])[0].get_measure(measure_reference)

    def get_identifier_references(self) -> List[IdentifierReference]:  # noqa: D
        return list(self._identifier_ref_to_entity.keys())

    def _get_joinable_data_sources_by_identifier(
        self, data_source_name: str, identifier_spec: LinkableInstanceSpec
    ) -> List[str]:
        """Returns list of data sources that can be joined to `data_source_name` via `identifier_spec`"""
        res = []
        for name, links in self._data_source_links.adj(data_source_name).items():
            for link in links:
                if link.via_from.name.element_name == identifier_spec.element_name:
                    res.append(name)
                    break

        return res

    def _get_joinable_data_sources(self, data_source_name: str) -> Dict[str, List[JoinLink]]:  # noqa: D
        return self._data_source_links.adj(data_source_name)

    # DSC interface
    def get_data_sources_for_measure(self, measure_reference: MeasureReference) -> List[DataSource]:  # noqa: D
        return self._measure_index[measure_reference]

    def dimension_is_partitioned(self, dimension_reference: DimensionReference) -> bool:  # noqa: D
        return self.get_dimension(dimension_reference).is_partition

    def keys(self) -> List[str]:  # noqa: D
        return list(self._data_source_names)

    def get_data_source_element(self, ref: DataSourceElementReference) -> Optional[Element]:  # Noqa: d
        data_source = self[ref.data_source_name]
        for elem in data_source.elements:
            if elem.name.element_name == ref.element_name:
                return elem

        return None

    def __getitem__(self, item: str) -> DataSource:  # noqa: D
        res = self.get(item)
        if res:
            return res

        raise ValueError(f"Cannot find data source with name ({item})")

    def get(self, data_source_name: str) -> Optional[DataSource]:  # noqa: D
        if data_source_name in self._configured_data_source_container:
            data_source = self._configured_data_source_container.get(data_source_name)
            assert isinstance(data_source, DataSource)
            return data_source

        return None

    def get_by_reference(self, data_source_reference: DataSourceReference) -> Optional[DataSource]:  # noqa: D
        return self.get(data_source_reference.data_source_name)

    def add_data_source(
        self,
        data_source: DataSource,
        fail_on_error: bool = True,
        logging_context: str = "",
    ) -> None:
        """Add data source semantic information, validating consistency with existing data sources and establishing links"""
        errors = []

        if data_source.name in self._data_source_names:
            errors.append(f"name {data_source.name} already registered - please ensure data source names are unique")

        if data_source.origin is DataSourceOrigin.DERIVED:
            raise ValueError(
                f"Cannot add derived data source (with name: {data_source.name}) to New DataSourceSemantics... for now"
            )

        for measure in data_source.measures:
            if measure.name in self._measure_aggs and self._measure_aggs[measure.name] != measure.agg:
                errors.append(
                    f"conflicting aggregation (agg) for measure `{measure.name}` registered as `{self._measure_aggs[measure.name]}`; "
                    f"Got `{measure.agg}"
                )

        if errors:
            error_prefix = "\n  - "
            error_msg = (
                f"Unable to add data source `{data_source.name}` "
                f"{'while ' + logging_context + ' ' if logging_context else ''}"
                f"{'... skipping' if not fail_on_error else ''}.\n"
                f"Errors: {error_prefix + error_prefix.join(errors)}"
            )
            if fail_on_error:
                raise InvalidDataSourceError(error_msg)
            logger.warning(error_msg)
            return

        self._data_source_names.add(data_source.name)

        for elem in data_source.elements:
            if elem.name not in self._element_types:
                self._element_types[elem.name] = type(elem)

        for meas in data_source.measures:
            self._measure_aggs[meas.name] = meas.agg
            self._measure_index[meas.name].append(data_source)
        for dim in data_source.dimensions:
            self._linkable_reference_index[dim.name].append(data_source)
            self._dimension_index[dim.name].append(data_source)
        for ident in data_source.identifiers:
            self._identifier_ref_to_entity[ident.name] = ident.entity
            self._entity_index[ident.entity].append(data_source)
            self._linkable_reference_index[ident.name].append(data_source)

        dsource_partition = data_source.partition
        # Add links to other data sources based on the names and types of identifiers available.
        for ident in data_source.identifiers:
            # We only need to consider other data sources with the same identifier.
            for other in self._entity_index[ident.entity]:
                if other.name == data_source.name:
                    continue

                # handle whether or not this join is partitioned
                partitions: List[DimensionReference] = list()
                other_partition = other.partition
                if dsource_partition and other_partition and dsource_partition.name != other_partition.name:
                    continue
                if dsource_partition and other_partition:
                    partitions.append(dsource_partition.name)

                for other_ident in other.identifiers:
                    # TODO: Replace with entity check when entities are supported.
                    if other_ident.name != ident.name:
                        continue

                    # add edge data_source -> other "other is joinable to data_source"
                    join_type = JOIN_TYPE_MAPPING.get((ident.type, other_ident.type), None)
                    if join_type:
                        self._data_source_links.add_edge(
                            from_key=data_source.name,
                            to_key=other.name,
                            value=JoinLink(
                                join_type=join_type,
                                via_from=ident,
                                via_to=other_ident,
                                partitions=tuple([str(p) for p in partitions]),
                            ),
                        )

                    # add edge other -> data_source "data_source is joinable to other"
                    reversed_join_type = JOIN_TYPE_MAPPING.get((other_ident.type, ident.type), None)
                    if reversed_join_type:
                        self._data_source_links.add_edge(
                            from_key=other.name,
                            to_key=data_source.name,
                            value=JoinLink(
                                join_type=reversed_join_type,
                                via_from=other_ident,
                                via_to=ident,
                                partitions=tuple([str(p) for p in partitions]),
                            ),
                        )

    @property
    def primary_time_dimension_reference(self) -> TimeDimensionReference:
        """Gets the primary time dimension that's used in all data sources."""
        for data_source in self._configured_data_source_container.values():
            for dimension in data_source.dimensions:
                if dimension.type == DimensionType.TIME and dimension.type_params and dimension.type_params.is_primary:
                    return TimeDimensionReference(element_name=dimension.name.element_name)
        raise RuntimeError("No primary time dimension found in the model")

    @property
    def data_source_references(self) -> Sequence[DataSourceReference]:  # noqa: D
        data_source_names_sorted = sorted(self._data_source_names)
        return tuple(DataSourceReference(data_source_name=x) for x in data_source_names_sorted)



import logging
from collections import defaultdict
from typing import List, Optional, Dict, Sequence

from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.elements.dimension import Dimension, DimensionType
from metricflow.specs import MeasureReference
from metricflow.time.time_constants import SUPPORTED_GRANULARITIES
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.validations.validator_helpers import (
    ModelValidationRule,
    ValidationIssue,
    ValidationIssueType,
    ValidationError,
    ValidationFatal,
    validate_safely,
)

logger = logging.getLogger(__name__)


class DataSourceMeasuresUniqueRule(ModelValidationRule):
    """Checks time dimensions in data sources."""

    @staticmethod
    @validate_safely(
        whats_being_done="running model validation ensuring measures exist in only one configured data source"
    )
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssueType]:  # noqa: D
        issues: List[ValidationIssueType] = []

        measure_names_to_data_sources: Dict[MeasureReference, List] = defaultdict(list)
        for data_source in model.data_sources:
            for measure in data_source.measures:
                if measure.name in measure_names_to_data_sources:
                    model_object_reference = ValidationIssue.make_object_reference(
                        data_source_name=data_source.name, measure_name=measure.name.element_name
                    )
                    issues.append(
                        ValidationError(
                            model_object_reference=model_object_reference,
                            message=f"Found measure with name {measure.name} in multiple data sources with names "
                            f"({measure_names_to_data_sources[measure.name]})",
                        )
                    )
                measure_names_to_data_sources[measure.name].append(data_source.name)

        return issues


class DataSourceTimeDimensionWarningsRule(ModelValidationRule):
    """Checks time dimensions in data sources."""

    @staticmethod
    @validate_safely(whats_being_done="running model validation ensuring time dimensions are defined properly")
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssueType]:  # noqa: D
        issues: List[ValidationIssueType] = []

        primary_time_dimension_name = None
        for data_source in model.data_sources:
            dimensions: Optional[Sequence[Dimension]] = data_source.dimensions
            for dimension in dimensions or []:
                if (
                    dimension.type == DimensionType.TIME
                    and dimension.type_params is not None
                    and dimension.type_params.is_primary is True
                ):
                    primary_time_dimension_name = dimension.name

        for data_source in model.data_sources:
            issues.extend(
                DataSourceTimeDimensionWarningsRule._validate_data_source(
                    data_source=data_source, primary_time_dimension_name=primary_time_dimension_name
                )
            )
        return issues

    @staticmethod
    @validate_safely(whats_being_done="checking validity of the data source's time dimensions")
    def _validate_data_source(
        data_source: DataSource, primary_time_dimension_name: Optional[str] = None
    ) -> List[ValidationIssueType]:
        primary_time_present = False
        issues: List[ValidationIssueType] = []

        for dim in data_source.dimensions:
            model_object_reference = ValidationIssue.make_object_reference(
                data_source_name=data_source.name, dimension_name=dim.name.element_name
            )

            if dim.type == DimensionType.TIME:
                if dim.type_params is None:
                    continue
                elif dim.type_params.is_primary:
                    primary_time_present = True
                    if primary_time_dimension_name and dim.name != primary_time_dimension_name:
                        issues.append(
                            ValidationFatal(
                                model_object_reference=model_object_reference,
                                message=f"In data source {data_source.name}, "
                                f"found primary time dimension with name: {dim.name}, "
                                f"another primary time dimension with name: {primary_time_dimension_name} "
                                f"was already found in model.",
                            )
                        )
                elif dim.type_params.time_granularity:
                    if dim.type_params.time_granularity not in SUPPORTED_GRANULARITIES:
                        issues.append(
                            ValidationError(
                                model_object_reference=model_object_reference,
                                message=f"Unsupported time granularity in time dimension with name: {dim.name}, "
                                f"Please use {[s.value for s in SUPPORTED_GRANULARITIES]}",
                            )
                        )

        if not primary_time_present and len(data_source.measures) > 0:
            issues.append(
                ValidationError(
                    model_object_reference=ValidationIssue.make_object_reference(data_source_name=data_source.name),
                    message=f"No primary time dimension in data source with name ({data_source.name}). Please add one",
                )
            )

        return issues



import inspect
import logging
import os.path
from dataclasses import dataclass
from string import Template
from typing import Optional, Dict, List, Union, Type, Any, Tuple

import yaml
from jsonschema import exceptions
from yaml.scanner import ScannerError

from metricflow.errors.errors import ParsingException
from metricflow.model.model_transformer import ModelTransformer
from metricflow.model.objects.common import Version
from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.materialization import Materialization
from metricflow.model.objects.metric import Metric
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.objects.utils import ParseableObject, ParseableField
from metricflow.model.parsing.validation import (
    validate_config_structure,
    VERSION_KEY,
    METRIC_TYPE,
    DATA_SOURCE_TYPE,
    MATERIALIZATION_TYPE,
    DOCUMENT_TYPES,
    DERIVED_GROUP_BY_ELEMENT_TYPE,
)
from metricflow.model.parsing.yaml_loader import ParsingContext, SafeLineLoader, PARSING_CONTEXT_KEY
from metricflow.model.validations.validator_helpers import ValidationIssueType

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class YamlFile:  # noqa: D
    file_path: str
    contents: str


@dataclass(frozen=True)
class ModelBuildResult:  # noqa: D
    model: Optional[UserConfiguredModel] = None
    # Issues found in the model.
    issues: Optional[Tuple[ValidationIssueType, ...]] = None


def parse_directory_of_yaml_files_to_model(
    directory: str,
    template_mapping: Optional[Dict[str, str]] = None,
    apply_transformations: bool = True,
    apply_pre_transformations: Optional[bool] = False,
    apply_post_transformations: Optional[bool] = False,
) -> ModelBuildResult:
    """Parse files in the given directory to a TMdoModel.

    Strings in the file following the Python string template format are replaced according to the template_mapping dict.
    """
    template_mapping = template_mapping or {}
    yaml_config_files = []

    for root, dirs, files in os.walk(directory):
        for file in files:
            file_path = os.path.join(root, file)
            with open(file_path) as f:
                contents = Template(f.read()).substitute(template_mapping)
                yaml_config_files.append(
                    YamlFile(file_path=file_path, contents=contents),
                )
    model = parse_yaml_files_to_model(yaml_config_files)
    if apply_transformations:
        model = ModelTransformer.transform_model(model)

    if apply_pre_transformations:
        model = ModelTransformer.pre_validation_transform_model(model)

    if apply_post_transformations:
        model = ModelTransformer.post_validation_transform_model(model)

    return ModelBuildResult(model=model)


def parse_yaml_files_to_model(files: List[YamlFile]) -> UserConfiguredModel:
    """Builds UserConfiguredModel from list of config files (as strings).

    Persistent storage connection may be passed to write parsed objects=
    to storage and populate object metadata

    Note: this function does not finalize the model
    """
    data_sources = []
    metrics = []
    materializations: List[Materialization] = []
    for config_file in files:
        objects = parse_config_yaml(config_file)  # parse config file
        for obj in objects:
            if isinstance(obj, DataSource):
                data_sources.append(obj)
            elif isinstance(obj, Metric):
                metrics.append(obj)
            elif isinstance(obj, Materialization):
                materializations.append(obj)
            else:
                raise ValueError(
                    f"Unexpected model object parsed from file at path: {config_file.file_path}. "
                    f"Expected internal versions of Metric, DataSource, Materialization, or DerivedGroupByElement"
                )

    return UserConfiguredModel(
        data_sources=data_sources,
        materializations=materializations,
        metrics=metrics,
    )


def parse_config_yaml(
    config_yaml: YamlFile,
) -> List[Union[DataSource, Metric, Materialization]]:
    """Parses transform config file passed as string - Returns list of model objects"""
    results: List[Union[DataSource, Metric, Materialization]] = []
    filename: str = os.path.split(config_yaml.file_path)[-1]
    ctx: Optional[ParsingContext] = None
    errors = []
    try:
        # Validates that config yaml conforms to json schema
        validate_config_structure(config_yaml.contents)

        for config_document in yaml.load_all(config_yaml.contents, Loader=SafeLineLoader):
            # The config document can be None if there is nothing but white space between two `---`
            # this isn't really an issue, so lets just swallow it
            if config_document is None:
                continue
            if not isinstance(config_document, dict):
                errors.append(
                    str(
                        ParsingException(
                            f"Transform object YAML must be a dict. Got `{type(config_document)}`: {config_document}",
                        )
                    )
                )
                continue
            keys = config_document.keys()
            if PARSING_CONTEXT_KEY not in keys:
                raise RuntimeError(f"No parsing ctx present, expected key `{PARSING_CONTEXT_KEY}`. Cannot parse yaml.")

            ctx = config_document.pop(PARSING_CONTEXT_KEY)
            assert ctx
            ctx.filename = filename

            if VERSION_KEY in config_document:
                version = Version.parse(config_document.pop(VERSION_KEY))
                major_version = version.major

                if major_version != 0:
                    raise RuntimeError(f"Unsupported version {version} in config document")

            if len(keys) != 1:
                errors.append(
                    str(
                        ParsingException(
                            f"Transform config document should have one type of key, but this has {len(keys)}. "
                            f"Found keys: {keys} in {config_yaml.file_path}",
                            ctx,
                        )
                    )
                )
                continue

            # retrieve last top-level key as type
            document_type = next(iter(config_document.keys()))
            object_cfg = config_document[document_type]
            if document_type == METRIC_TYPE:
                results.append(parse(Metric, ctx, object_cfg, config_yaml.file_path, config_yaml.contents))
            elif document_type == DATA_SOURCE_TYPE:
                results.append(parse(DataSource, ctx, object_cfg, config_yaml.file_path, config_yaml.contents))
            elif document_type == MATERIALIZATION_TYPE:
                results.append(parse(Materialization, ctx, object_cfg, config_yaml.file_path, config_yaml.contents))
            elif document_type == DERIVED_GROUP_BY_ELEMENT_TYPE:
                logger.warning(f"{DERIVED_GROUP_BY_ELEMENT_TYPE} is not supported in {config_yaml.file_path}")
            else:
                errors.append(
                    str(
                        ParsingException(
                            message=f"Invalid document type: {document_type}. Expected {DOCUMENT_TYPES}",
                            ctx=ctx,
                        )
                    )
                )

        if len(errors) > 0:
            raise ParsingException(message=f"Found {len(errors)} errors parsing configs. Errors: {errors}")
    except exceptions.ValidationError as e:
        logger.exception(f"Config yaml did not conform to metric spec\nError: {e}")
        raise ParsingException(
            message=f"Config yaml did not conform to metric spec\nError: {e}",
            ctx=ctx,
        ) from e
    except ScannerError as e:
        logger.exception(f"Failed parsing invalid yaml.\nParsing context: {ctx}")
        raise ParsingException(message=f"Failed to parse yaml in file. Error: {e}", ctx=ctx) from e
    except Exception as e:
        logger.exception(f"Failed parsing yaml.\nParsing context: {ctx}")
        raise ParsingException(
            message=f"Failed to parse yaml in file. Error: {e}",
            ctx=ctx,
        ) from e

    return results


def parse(  # type: ignore[misc]
    _type: Type[Union[DataSource, Metric, Materialization]],
    ctx: ParsingContext,
    yaml_dict: Dict[str, Any],
    filename: str,
    contents: str,
) -> Any:
    """Parses a model object from (jsonschema-validated) yaml into python object"""

    #  Only - add MdoMetadata
    yaml_dict["metadata"] = {
        "repo_file_path": filename,
        "file_slice": {
            "filename": ctx.filename,
            "content": contents,
            "start_line_number": ctx.start_line,
            "end_line_number": ctx.end_line,
        },
    }

    for field_name, field_value in _type.__fields__.items():
        if field_name in yaml_dict:
            if not inspect.isclass(field_value.type_):  # this handles the nested generic-type case (eg List[List[str]])
                continue
            if issubclass(field_value.type_, ParseableObject):
                if isinstance(yaml_dict[field_name], list):
                    objects = []
                    for obj in yaml_dict[field_name]:
                        objects.append(parse(field_value.type_, ctx, obj, filename, contents))  # type: ignore
                    yaml_dict[field_name] = objects
                else:
                    yaml_dict[field_name] = parse(field_value.type_, ctx, yaml_dict[field_name], filename, contents)  # type: ignore
            elif issubclass(field_value.type_, ParseableField):
                if isinstance(yaml_dict[field_name], list):
                    objects = []
                    for obj in yaml_dict[field_name]:
                        objects.append(field_value.type_.parse(obj))
                    yaml_dict[field_name] = objects
                else:
                    yaml_dict[field_name] = field_value.type_.parse(yaml_dict[field_name])

    return _type(**yaml_dict)



import logging
import re
from dataclasses import dataclass
from typing import Optional, List, Tuple, Sequence

from metricflow.dag.id_generation import IdGeneratorRegistry
from metricflow.dataflow.sql_table import SqlTable
from metricflow.instances import (
    MeasureInstance,
    DataSourceElementReference,
    DimensionInstance,
    TimeDimensionInstance,
    IdentifierInstance,
    InstanceSet,
    DataSourceReference,
    AggregationState,
)
from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.elements.dimension import Dimension, DimensionType
from metricflow.model.objects.elements.identifier import Identifier, IdentifierType
from metricflow.model.objects.elements.measure import Measure
from metricflow.specs import (
    TimeDimensionSpec,
    MeasureSpec,
    DimensionSpec,
    IdentifierSpec,
    ColumnAssociationResolver,
    LinklessIdentifierSpec,
)
from metricflow.sql.sql_exprs import (
    SqlStringExpression,
    SqlDateTruncExpression,
    SqlColumnReferenceExpression,
    SqlColumnReference,
    SqlExpressionNode,
)
from metricflow.sql.sql_plan import (
    SqlTableFromClauseNode,
    SqlSelectStatementNode,
    SqlSelectColumn,
    SqlQueryPlanNode,
    SqlSelectQueryFromClauseNode,
)
from metricflow.time.time_granularity import TimeGranularity
from metricflow.dataset.data_source_adapter import DataSourceDataSet

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class DimensionConversionResult:
    """Helper class for returning the results of converting dimension from a data source."""

    dimension_instances: Sequence[DimensionInstance]
    time_dimension_instances: Sequence[TimeDimensionInstance]
    select_columns: Sequence[SqlSelectColumn]


class DataSourceToDataSetConverter:
    """Converts a data source in the old MQL model to the data set that can be used with the dataflow plan builder."""

    # Regex for inferring whether an expression for an element is a column reference.
    _SQL_IDENTIFIER_REGEX = re.compile("^[a-zA-Z_][a-zA-Z_0-9]*$")

    def __init__(self, column_association_resolver: ColumnAssociationResolver) -> None:  # noqa: D
        self._column_association_resolver = column_association_resolver

    def _create_dimension_instance(
        self,
        data_source_name: str,
        dimension: Dimension,
        identifier_links: Tuple[LinklessIdentifierSpec, ...],
    ) -> DimensionInstance:
        """Create a dimension instance from the old MQL dimension type."""
        dimension_spec = DimensionSpec(
            element_name=dimension.name.element_name,
            identifier_links=identifier_links,
        )
        column_associations = dimension_spec.column_associations(self._column_association_resolver)

        return DimensionInstance(
            associated_columns=column_associations,
            spec=dimension_spec,
            defined_from=(
                DataSourceElementReference(
                    data_source_name=data_source_name,
                    element_name=dimension.name.element_name,
                ),
            ),
        )

    def _create_time_dimension_instance(
        self,
        data_source_name: str,
        time_dimension: Dimension,
        identifier_links: Tuple[LinklessIdentifierSpec, ...],
        time_granularity: Optional[TimeGranularity] = None,
    ) -> TimeDimensionInstance:
        """Create a time dimension instance from the old MQL dimension type."""
        time_dimension_spec = TimeDimensionSpec(
            element_name=time_dimension.name.element_name,
            identifier_links=identifier_links,
            time_granularity=time_granularity,
        )

        column_associations = time_dimension_spec.column_associations(self._column_association_resolver)

        return TimeDimensionInstance(
            associated_columns=column_associations,
            spec=time_dimension_spec,
            defined_from=(
                DataSourceElementReference(
                    data_source_name=data_source_name,
                    element_name=time_dimension.name.element_name,
                ),
            ),
        )

    def _create_identifier_instance(
        self,
        data_source_name: str,
        identifier: Identifier,
        identifier_links: Tuple[LinklessIdentifierSpec, ...],
    ) -> IdentifierInstance:
        """Create a dimension instance from the old MQL dimension type."""
        identifier_spec = IdentifierSpec(
            element_name=identifier.name.element_name,
            identifier_links=identifier_links,
        )
        column_associations = identifier_spec.column_associations(self._column_association_resolver)

        return IdentifierInstance(
            associated_columns=column_associations,
            spec=identifier_spec,
            defined_from=(
                DataSourceElementReference(
                    data_source_name=data_source_name,
                    element_name=identifier.name.element_name,
                ),
            ),
        )

    @staticmethod
    def _make_element_sql_expr(
        table_alias: str, element_name: str, element_expr: Optional[str] = None
    ) -> SqlExpressionNode:
        """Create an expression that can be used for reading the element from the data source's SQL."""
        if element_expr:
            if DataSourceToDataSetConverter._SQL_IDENTIFIER_REGEX.match(element_expr) and element_expr.upper() not in (
                "TRUE",
                "FALSE",
                "NULL",
            ):
                return SqlColumnReferenceExpression(
                    SqlColumnReference(
                        table_alias=table_alias,
                        column_name=element_expr,
                    )
                )
            return SqlStringExpression(sql_expr=element_expr)

        return SqlColumnReferenceExpression(
            SqlColumnReference(
                table_alias=table_alias,
                column_name=element_name,
            )
        )

    def _convert_measures(
        self,
        data_source_name: str,
        measures: Sequence[Measure],
        measure_time_dimension_spec: TimeDimensionSpec,
        table_alias: str,
    ) -> Tuple[Sequence[MeasureInstance], Sequence[SqlSelectColumn]]:
        # Convert all elements to instances
        measure_instances = []
        select_columns = []
        for measure in measures or []:
            measure_spec = MeasureSpec(
                element_name=measure.name.element_name,
            )
            measure_instance = MeasureInstance(
                associated_columns=measure_spec.column_associations(self._column_association_resolver),
                source_time_dimension_reference=measure_time_dimension_spec.reference,
                spec=measure_spec,
                defined_from=(
                    DataSourceElementReference(
                        data_source_name=data_source_name,
                        element_name=measure.name.element_name,
                    ),
                ),
                aggregation_state=AggregationState.NON_AGGREGATED,
            )
            measure_instances.append(measure_instance)
            select_columns.append(
                SqlSelectColumn(
                    expr=DataSourceToDataSetConverter._make_element_sql_expr(
                        table_alias=table_alias,
                        element_name=measure.name.element_name,
                        element_expr=measure.expr,
                    ),
                    column_alias=measure_instance.associated_column.column_name,
                )
            )

        return measure_instances, select_columns

    def _convert_dimensions(
        self,
        data_source_name: str,
        dimensions: Sequence[Dimension],
        identifier_links: Tuple[LinklessIdentifierSpec, ...],
        table_alias: str,
    ) -> DimensionConversionResult:
        dimension_instances = []
        time_dimension_instances = []
        select_columns = []

        for dimension in dimensions or []:
            if dimension.type == DimensionType.CATEGORICAL:
                dimension_instance = self._create_dimension_instance(
                    data_source_name=data_source_name,
                    dimension=dimension,
                    identifier_links=identifier_links,
                )
                dimension_instances.append(dimension_instance)
                select_columns.append(
                    SqlSelectColumn(
                        expr=DataSourceToDataSetConverter._make_element_sql_expr(
                            table_alias=table_alias,
                            element_name=dimension.name.element_name,
                            element_expr=dimension.expr,
                        ),
                        column_alias=dimension_instance.associated_column.column_name,
                    )
                )

            elif dimension.type == DimensionType.TIME:
                defined_time_granularity = TimeGranularity.DAY
                if dimension.type_params and dimension.type_params.time_granularity:
                    defined_time_granularity = dimension.type_params.time_granularity

                time_dimension_instance = self._create_time_dimension_instance(
                    data_source_name=data_source_name,
                    time_dimension=dimension,
                    identifier_links=identifier_links,
                    time_granularity=defined_time_granularity,
                )
                time_dimension_instances.append(time_dimension_instance)
                select_columns.append(
                    SqlSelectColumn(
                        expr=DataSourceToDataSetConverter._make_element_sql_expr(
                            table_alias=table_alias,
                            element_name=dimension.name.element_name,
                            element_expr=dimension.expr,
                        ),
                        column_alias=time_dimension_instance.associated_column.column_name,
                    )
                )

                # Add time dimensions with a smaller granularity for ease in query resolution
                for time_granularity in TimeGranularity:
                    if time_granularity.to_int() > defined_time_granularity.to_int():
                        time_dimension_instance = self._create_time_dimension_instance(
                            data_source_name=data_source_name,
                            time_dimension=dimension,
                            identifier_links=identifier_links,
                            time_granularity=time_granularity,
                        )
                        time_dimension_instances.append(time_dimension_instance)

                        select_columns.append(
                            SqlSelectColumn(
                                expr=SqlDateTruncExpression(
                                    time_granularity=time_granularity,
                                    arg=DataSourceToDataSetConverter._make_element_sql_expr(
                                        table_alias=table_alias,
                                        element_name=dimension.name.element_name,
                                        element_expr=dimension.expr,
                                    ),
                                ),
                                column_alias=time_dimension_instance.associated_column.column_name,
                            )
                        )
            else:
                assert False, f"Unhandled dimension type: {dimension.type}"

        return DimensionConversionResult(
            dimension_instances=dimension_instances,
            time_dimension_instances=time_dimension_instances,
            select_columns=select_columns,
        )

    def _create_identifier_instances(
        self,
        data_source_name: str,
        identifiers: Sequence[Identifier],
        identifier_links: Tuple[LinklessIdentifierSpec, ...],
        table_alias: str,
    ) -> Tuple[Sequence[IdentifierInstance], Sequence[SqlSelectColumn]]:
        identifier_instances = []
        select_columns = []
        for identifier in identifiers or []:
            # We don't want to create something like user_id__user_id, so skip if the link is the same as the
            # identifier.
            if (
                len(identifier_links) == 1
                and LinklessIdentifierSpec.from_element_name(identifier.name.element_name) == identifier_links[0]
            ):
                continue

            identifier_instance = self._create_identifier_instance(
                data_source_name=data_source_name,
                identifier=identifier,
                identifier_links=identifier_links,
            )

            identifier_instances.append(identifier_instance)
            if identifier.is_composite:
                for idx in range(len(identifier.identifiers)):
                    sub_id = identifier.identifiers[idx]
                    column_name = identifier_instance.associated_columns[idx].column_name

                    expr = sub_id.expr
                    if expr is None:
                        assert sub_id.name is not None
                        expr = sub_id.name.element_name
                    sub_id_name = sub_id.ref or sub_id.name
                    assert sub_id_name, f"Sub-identifier {sub_id} must have 'name' or 'ref' defined"

                    select_columns.append(
                        SqlSelectColumn(
                            expr=DataSourceToDataSetConverter._make_element_sql_expr(
                                table_alias=table_alias,
                                element_name=sub_id_name,
                                element_expr=expr,
                            ),
                            column_alias=column_name,
                        )
                    )
            else:
                select_columns.append(
                    SqlSelectColumn(
                        expr=DataSourceToDataSetConverter._make_element_sql_expr(
                            table_alias=table_alias,
                            element_name=identifier.name.element_name,
                            element_expr=identifier.expr,
                        ),
                        column_alias=identifier_instance.associated_column.column_name,
                    )
                )
        return identifier_instances, select_columns

    @staticmethod
    def _find_primary_time_dimension(data_source: DataSource) -> TimeDimensionSpec:
        # If a data source has measures, it should have a primary time dimension. Find it and use it to set the time
        # dimension for the measure.
        primary_time_dimensions = [dimension for dimension in data_source.dimensions if dimension.is_primary_time]
        assert len(primary_time_dimensions) == 1, (
            f"Data source ({data_source}) with measures should have exactly 1 primary time dimension, found "
            f"{len(primary_time_dimensions)}: {primary_time_dimensions}."
        )
        primary_time_dimension = primary_time_dimensions[0]
        assert (
            primary_time_dimension.type_params and primary_time_dimension.type_params.time_granularity
        ), f"Primary time dimension missing time granularity: {primary_time_dimension}"

        return TimeDimensionSpec(
            element_name=primary_time_dimension.name.element_name,
            identifier_links=(),
            time_granularity=primary_time_dimension.type_params.time_granularity,
        )

    def create_sql_source_data_set(self, data_source: DataSource) -> DataSourceDataSet:
        """Create an SQL source data set from the old MQL configured data sources."""

        # Gather all instances and columns from all data sources.
        all_measure_instances: List[MeasureInstance] = []
        all_dimension_instances: List[DimensionInstance] = []
        all_time_dimension_instances: List[TimeDimensionInstance] = []
        all_identifier_instances: List[IdentifierInstance] = []

        all_select_columns: List[SqlSelectColumn] = []
        from_source_alias = IdGeneratorRegistry.for_class(self.__class__).create_id(f"{data_source.name}_src")

        # Handle measures
        if len(data_source.measures) > 0:
            primary_time_dimension = self._find_primary_time_dimension(data_source)
            measure_instances, select_columns = self._convert_measures(
                data_source_name=data_source.name,
                measures=data_source.measures,
                measure_time_dimension_spec=primary_time_dimension,
                table_alias=from_source_alias,
            )
            all_measure_instances.extend(measure_instances)
            all_select_columns.extend(select_columns)

        # For dimensions in a data source, we can access them through the local form, or the dundered form.
        # e.g. in the "users" data source, with the "country" dimension and the "user_id" identifier,
        # the dimensions "country" and "user_id__country" both mean the same thing. To make matching easier, create both
        # instances in the instance set. We'll create a different instance for each "possible_identifier_links".
        possible_identifier_links: List[Tuple[LinklessIdentifierSpec, ...]] = [()]
        for identifier in data_source.identifiers:
            if identifier.type in (IdentifierType.PRIMARY, IdentifierType.UNIQUE):
                possible_identifier_links.append(
                    (LinklessIdentifierSpec.from_element_name(element_name=identifier.name.element_name),)
                )

        # Handle dimensions
        conversion_results = [
            self._convert_dimensions(
                data_source_name=data_source.name,
                dimensions=data_source.dimensions,
                identifier_links=identifier_links,
                table_alias=from_source_alias,
            )
            for identifier_links in possible_identifier_links
        ]

        all_dimension_instances.extend(
            [
                dimension_instance
                for conversion_result in conversion_results
                for dimension_instance in conversion_result.dimension_instances
            ]
        )

        all_time_dimension_instances.extend(
            [
                time_dimension_instance
                for conversion_result in conversion_results
                for time_dimension_instance in conversion_result.time_dimension_instances
            ]
        )

        all_select_columns.extend(
            [
                select_column
                for conversion_result in conversion_results
                for select_column in conversion_result.select_columns
            ]
        )

        # Handle identifiers
        for identifier_links in possible_identifier_links:
            identifier_instances, select_columns = self._create_identifier_instances(
                data_source_name=data_source.name,
                identifiers=data_source.identifiers,
                identifier_links=identifier_links,
                table_alias=from_source_alias,
            )
            all_identifier_instances.extend(identifier_instances)
            all_select_columns.extend(select_columns)

        # Generate the "from" clause depending on whether it's an SQL query or an SQL table.
        from_source: Optional[SqlQueryPlanNode] = None
        if data_source.sql_table:
            from_source = SqlTableFromClauseNode(sql_table=SqlTable.from_string(data_source.sql_table))
        elif data_source.sql_query:
            from_source = SqlSelectQueryFromClauseNode(select_query=data_source.sql_query)
        else:
            raise RuntimeError(f"Data source does not have sql_table or sql_query set: {data_source}")

        select_statement_node = SqlSelectStatementNode(
            description=f"Read elements from data source '{data_source.name}'",
            select_columns=tuple(all_select_columns),
            from_source=from_source,
            from_source_alias=from_source_alias,
            joins_descs=(),
            where=None,
            group_bys=(),
            order_bys=(),
        )

        return DataSourceDataSet(
            data_source_reference=DataSourceReference(data_source_name=data_source.name),
            instance_set=InstanceSet(
                measure_instances=tuple(all_measure_instances),
                dimension_instances=tuple(all_dimension_instances),
                time_dimension_instances=tuple(all_time_dimension_instances),
                identifier_instances=tuple(all_identifier_instances),
                metric_instances=(),
            ),
            sql_select_node=select_statement_node,
        )



import logging
import textwrap
from typing import ClassVar, Optional, List, Set

import jinja2
import sqlalchemy

from metricflow.dataflow.sql_table import SqlTable
from metricflow.protocols.sql_client import SupportedSqlEngine, SqlEngineAttributes
from metricflow.sql.render.sql_plan_renderer import SqlQueryPlanRenderer
from metricflow.sql.render.sqlite import SqliteSqlQueryPlanRenderer
from metricflow.sql_clients.common_client import SqlDialect
from metricflow.sql_clients.sqlalchemy_dialect import SqlAlchemySqlClient


class SQLiteEngineAttributes:
    """Engine-specific attributes for the SQLite query engine

    This is an implementation of the SqlEngineAttributes protocol for SQLite
    """

    sql_engine_type: ClassVar[SupportedSqlEngine] = SupportedSqlEngine.SQLITE

    # SQL Engine capabilities
    date_trunc_supported: ClassVar[bool] = False
    full_outer_joins_supported: ClassVar[bool] = False
    indexes_supported: ClassVar[bool] = True
    multi_threading_supported: ClassVar[bool] = False
    timestamp_type_supported: ClassVar[bool] = False
    timestamp_to_string_comparison_supported: ClassVar[bool] = False

    # SQL Dialect replacement strings
    double_data_type_name: ClassVar[str] = "DOUBLE"
    timestamp_type_name: ClassVar[Optional[str]] = None

    # MetricFlow attributes
    sql_query_plan_renderer: ClassVar[SqlQueryPlanRenderer] = SqliteSqlQueryPlanRenderer()


# TODO: add implementation of SQLite-compatible SqlClient
logger = logging.getLogger(__name__)


class SqliteSqlClient(SqlAlchemySqlClient):
    """Generally useful for testing ... not recommended for persistent use cases"""

    # Name of the schema created automatically.
    DEFAULT_DB = "main"

    @staticmethod
    def from_connection_details(url: str, password: Optional[str] = None) -> SqlAlchemySqlClient:  # noqa: D
        expected_url = "sqlite://"
        if url != expected_url:
            raise ValueError(f"URL was '{url}' but should be '{expected_url}'")
        if password:
            raise ValueError("Password should be empty")

        return SqliteSqlClient()

    def __init__(self) -> None:  # noqa: D
        super().__init__(sqlalchemy.create_engine(f"{SqlDialect.SQLITE.value}:///:memory:", pool_pre_ping=True))
        self._created_schemas: Set[str] = {self.DEFAULT_DB}

    @property
    def sql_engine_attributes(self) -> SqlEngineAttributes:
        """Collection of attributes and features specific to the SQLite SQL engine"""
        return SQLiteEngineAttributes()

    def create_schema(self, schema_name: str) -> None:  # noqa: D
        if schema_name not in self._created_schemas:
            self.execute(
                stmt=f"ATTACH DATABASE '/tmp/{schema_name}.db' AS {schema_name}",
            )
            self._created_schemas.add(schema_name)
        else:
            logger.info(f"create_schema() called with schema that exists '{schema_name}', so not actually creating it.")

    def list_tables(self, schema_name: str) -> List[str]:  # noqa:D
        return self.query(
            f"""\
            SELECT
              name
            FROM {schema_name}.sqlite_master
            WHERE type = 'table'
            """,
        )["name"].tolist()

    def table_exists(self, sql_table: SqlTable) -> bool:  # noqa :D
        return self.query(
            jinja2.Template(
                textwrap.dedent(
                    """\
                    SELECT EXISTS (
                      SELECT name FROM {{ schema_name }}.sqlite_master
                      WHERE type='table' AND name = '{{ table_name }}'
                    )
                    """,
                ),
                undefined=jinja2.StrictUndefined,
            ).render(
                schema_name=sql_table.schema_name,
                table_name=sql_table.table_name,
            )
        ).squeeze()

    def drop_schema(self, schema_name: str, cascade: bool = True) -> None:  # noqa: D
        logger.info(f"Not dropping schema '{schema_name}' since this SQLite DB is ephemeral")



from __future__ import annotations
from typing import Optional

from metricflow.model.objects.utils import FrozenBaseModel, ParseableField


class SqlTable(FrozenBaseModel, ParseableField):
    """Represents a reference to a SQL table."""

    db_name: Optional[str] = None
    schema_name: str
    table_name: str

    @staticmethod
    def parse(s: str) -> SqlTable:
        """Implement ParseableField interface"""
        return SqlTable.from_string(s)

    @staticmethod
    def from_string(sql_str: str) -> SqlTable:  # noqa: D
        sql_str_split = sql_str.split(".")
        if len(sql_str_split) == 2:
            return SqlTable(schema_name=sql_str_split[0], table_name=sql_str_split[1])
        elif len(sql_str_split) == 3:
            return SqlTable(db_name=sql_str_split[0], schema_name=sql_str_split[1], table_name=sql_str_split[2])
        raise RuntimeError(f"Invalid input for a SQL table, expected form '<schema>.<table>' but got: {sql_str}")

    @property
    def sql(self) -> str:
        """Return the snippet that can be used for use in SQL queries."""
        if self.db_name:
            return f"{self.db_name}.{self.schema_name}.{self.table_name}"
        return f"{self.schema_name}.{self.table_name}"



from metricflow.sql.render.expr_renderer import (
    DefaultSqlExpressionRenderer,
    SqlExpressionRenderer,
    SqlExpressionRenderResult,
    SqlTimeDeltaExpression,
)
from metricflow.sql.render.sql_plan_renderer import DefaultSqlQueryPlanRenderer
from metricflow.sql.sql_exprs import SqlCastToTimestampExpression, SqlDateTruncExpression


class SqliteSqlExpressionRenderer(DefaultSqlExpressionRenderer):
    """Expression renderer for the SQLite engine."""

    def visit_cast_to_timestamp_expr(self, node: SqlCastToTimestampExpression) -> SqlExpressionRenderResult:  # noqa: D
        # SQLite doesn't support timestamps, so all datetime related values are treated as strings.
        arg_rendered = self.render_sql_expr(node.arg)
        return SqlExpressionRenderResult(
            sql=f"CAST({arg_rendered.sql} AS TEXT)",
            execution_parameters=arg_rendered.execution_parameters,
        )

    def visit_time_delta_expr(self, node: SqlTimeDeltaExpression) -> SqlExpressionRenderResult:  # noqa: D
        arg_rendered = node.arg.accept(self)
        if node.grain_to_date:
            return SqlExpressionRenderResult(
                sql=f"DATE({arg_rendered.sql}, 'start of {node.granularity.value}')",
                execution_parameters=arg_rendered.execution_parameters,
            )

        return SqlExpressionRenderResult(
            sql=f"DATE({arg_rendered.sql}, '-{node.count} {node.granularity.value}')",
            execution_parameters=arg_rendered.execution_parameters,
        )

    def visit_date_trunc_expr(self, node: SqlDateTruncExpression) -> SqlExpressionRenderResult:  # noqa: D
        arg_rendered = self.render_sql_expr(node.arg)

        # SQLite doesn't have the method, but it would be helpful to show something in the plan outputs.
        return SqlExpressionRenderResult(
            sql="'__DATE_TRUNC_NOT_SUPPORTED__'",
            execution_parameters=arg_rendered.execution_parameters,
        )


class SqliteSqlQueryPlanRenderer(DefaultSqlQueryPlanRenderer):
    """Plan renderer for the SQLite engine."""

    EXPR_RENDERER = SqliteSqlExpressionRenderer()

    @property
    def expr_renderer(self) -> SqlExpressionRenderer:  # noqa :D
        return self.EXPR_RENDERER



import logging
import pprint
from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, List

from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.elements.identifier import Identifier, IdentifierType
from metricflow.model.objects.metric import Metric
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.validations.validator_helpers import ModelObjectType
from metricflow.specs import (
    MeasureReference,
    DimensionReference,
    IdentifierReference,
    IdentifierSpec,
    DimensionSpec,
    LinklessIdentifierSpec,
)

logger = logging.getLogger(__name__)
pp = pprint.PrettyPrinter(indent=2)


@dataclass(frozen=True)
class IdentifierKey:
    """Identifier used for keys in dicts"""

    name: str
    type: IdentifierType


class DimensionAndIdentifierNameValidator:
    """Checks if a given dimension or identifier name is valid for querying with a given metric"""

    def __init__(self, model: UserConfiguredModel):  # noqa: D
        # Map the metric name to the measures that it uses
        self._metric_name_to_measures: Dict[str, List[MeasureReference]] = {}
        metric: Metric
        for metric in model.metrics:
            measure_references: List[MeasureReference] = []
            if metric.type_params:
                if metric.type_params.measure:
                    measure_references.append(metric.type_params.measure)
                if metric.type_params.measures:
                    measure_references.extend(metric.type_params.measures)
                if metric.type_params.numerator:
                    measure_references.append(metric.type_params.numerator)
                if metric.type_params.denominator:
                    measure_references.append(metric.type_params.denominator)
            self._metric_name_to_measures[metric.name] = measure_references

        # Map the identifier name / type to the associated dimensions
        # Keys are (name of identifier, IdentifierType value)
        # Values are the list of dimension names available.
        self._identifier_key_to_dimension_names: Dict[IdentifierKey, List[str]] = defaultdict(list)
        data_source: DataSource
        for data_source in model.data_sources:
            dimension_names = [x.name.element_name for x in data_source.dimensions]
            identifier: Identifier
            for identifier in data_source.identifiers:
                identifier_key = IdentifierKey(identifier.name.element_name, identifier.type)
                self._identifier_key_to_dimension_names[identifier_key].extend(dimension_names)

        # Map the measure to the identifier name / type present in the data source where the measure is defined.
        # Keys are the name of the measure
        # Values are (name of identifier, IdentifierType value)
        self._measure_to_associated_identifier_keys: Dict[MeasureReference, List[IdentifierKey]] = defaultdict(list)
        for data_source in model.data_sources:
            measure_references = [x.name for x in data_source.measures]
            # List of identifiers and their types in the data source
            identifier_keys: List[IdentifierKey] = []
            for identifier in data_source.identifiers:
                identifier_keys.append(IdentifierKey(identifier.name.element_name, identifier.type))
            for measure_reference in measure_references:
                self._measure_to_associated_identifier_keys[measure_reference].extend(identifier_keys)

        # Maps the type of identifier associated with the measure to the valid types of other identifiers associated
        # with dimensions
        self._measure_identifier_type_to_dimension_identifier_types = {
            IdentifierType.PRIMARY: [IdentifierType.PRIMARY, IdentifierType.UNIQUE],
            IdentifierType.UNIQUE: [IdentifierType.PRIMARY, IdentifierType.UNIQUE],
            IdentifierType.FOREIGN: [IdentifierType.PRIMARY, IdentifierType.UNIQUE],
        }

        # Think about case where measures are defined in two different data sources but with a different set of
        # dimensions or identifiers. Like tests/integration/semantics/itest_constraints.py::test_split_measure

        # Populate the list of local dimension for each measure
        self._measure_to_local_dimension: Dict[MeasureReference, List[DimensionReference]] = defaultdict(list)
        for data_source in model.data_sources:
            local_dimension_references = [dimension.name for dimension in data_source.dimensions]
            for measure in data_source.measures:
                self._measure_to_local_dimension[measure.name].extend(local_dimension_references)

        # Populate the list of identifiers for each measure
        self._measures_to_identifiers: Dict[MeasureReference, List[IdentifierReference]] = defaultdict(list)
        for data_source in model.data_sources:
            identifier_references = [identifier.name for identifier in data_source.identifiers]
            for measure in data_source.measures:
                self._measures_to_identifiers[measure.name].extend(identifier_references)

        # Populate the name of the element to the type
        self._element_name_to_type: Dict[str, ModelObjectType] = {}
        for data_source in model.data_sources:
            for measure in data_source.measures:
                self._element_name_to_type[measure.name.element_name] = ModelObjectType.MEASURE
            for dimension in data_source.dimensions:
                self._element_name_to_type[dimension.name.element_name] = ModelObjectType.DIMENSION
            for identifier in data_source.identifiers:
                self._element_name_to_type[identifier.name.element_name] = ModelObjectType.IDENTIFIER

    def _is_identifier_valid_for_measure(self, measure_reference: MeasureReference, identifier_name: str) -> bool:
        assert measure_reference in self._measures_to_identifiers
        identifier_spec = IdentifierSpec.parse(identifier_name)
        return identifier_spec in self._measures_to_identifiers[measure_reference]

    def is_dimension_valid_for_measure(self, measure_reference: MeasureReference, dimension_name: str) -> bool:
        """Return true iff the given measure and dimension could be queried together."""

        logger.debug(
            f"Checking if dimension name `{dimension_name} is valid for measure `{measure_reference.element_name}` "
        )
        dimension_spec = DimensionSpec.parse(dimension_name)

        # If there's no identifier, it must be a local dimension, so check there.
        if not dimension_spec.identifier_links:
            assert measure_reference in self._measure_to_local_dimension
            dimension_reference = DimensionReference(element_name=dimension_spec.element_name)
            return dimension_reference in self._measure_to_local_dimension[measure_reference]

        # It's a joined dimension since it has an identifier in the name.
        assert measure_reference in self._measure_to_associated_identifier_keys

        # The measure has these identifiers associated with them.
        for identifier_key in self._measure_to_associated_identifier_keys[measure_reference]:
            logger.debug(
                f"Measure {measure_reference.element_name} is associated with the identifier {identifier_key.name} of "
                f"type {identifier_key.type}"
            )
            # List of identifier types that may allow a joined dimension
            dimension_identifier_types = self._measure_identifier_type_to_dimension_identifier_types[
                identifier_key.type
            ]
            for dimension_identifier_type in dimension_identifier_types:
                logger.debug(
                    f"Looking for dimensions associated with identifiers with name {identifier_key.name} "
                    f"and type {identifier_key.type}"
                )

                if identifier_key in self._identifier_key_to_dimension_names:
                    dimension_element_names_with_given_identifier_name_and_type = (
                        self._identifier_key_to_dimension_names[
                            IdentifierKey(identifier_key.name, dimension_identifier_type)
                        ]
                    )
                    logger.debug(
                        f"Dimensions in data sources that contain identifiers named {identifier_key.name} "
                        f"and type {dimension_identifier_type} are "
                        f"{pp.pformat(dimension_element_names_with_given_identifier_name_and_type)}"
                    )

                    for dimension_element_name in dimension_element_names_with_given_identifier_name_and_type:
                        if (
                            dimension_spec.element_name == dimension_element_name
                            and LinklessIdentifierSpec.from_element_name(identifier_key.name)
                            in dimension_spec.identifier_links
                        ):
                            return True
                else:
                    logger.debug("Did not find associated dimensions.")
        logger.debug(f"Dimension name `{dimension_name}` is not valid for measure `{measure_reference.element_name}`")
        return False

    def is_dimension_valid_for_metric(self, metric_name: str, dimension_name: str) -> bool:
        """Return true iff the given dimension is valid to query with the given metric."""

        assert metric_name in self._metric_name_to_measures
        return all(
            self.is_dimension_valid_for_measure(measure_spec, dimension_name)
            for measure_spec in self._metric_name_to_measures[metric_name]
        )

    def is_identifier_valid_for_metric(self, metric_name: str, identifier_name: str) -> bool:
        """Return true iff the given identifier is valid to query with the given metric."""
        assert len(self._metric_name_to_measures[metric_name]) >= 1
        return all(
            self._is_identifier_valid_for_measure(measure_spec, identifier_name)
            for measure_spec in self._metric_name_to_measures[metric_name]
        )



import copy
import logging

from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.transformations.boolean_measure import BooleanMeasureAggregationRule
from metricflow.model.transformations.identifiers import CompositeIdentifierExpressionRule
from metricflow.model.transformations.names import LowerCaseNamesRule
from metricflow.model.transformations.proxy_measure import CreateProxyMeasureRule

logger = logging.getLogger(__name__)


class ModelTransformer:
    """Helps to make transformations to a model for convenience.

    Generally used to make it more convenient for the user to develop their model.
    """

    @staticmethod
    def transform_model(model: UserConfiguredModel) -> UserConfiguredModel:
        """Transform a model according to configured rules."""
        model_copy = copy.deepcopy(model)
        for transform_rule in (
            LowerCaseNamesRule(),
            CreateProxyMeasureRule(),
            BooleanMeasureAggregationRule(),
            CompositeIdentifierExpressionRule(),
        ):
            model_copy = transform_rule.transform_model(model_copy)

        return model_copy

    @staticmethod
    def pre_validation_transform_model(model: UserConfiguredModel) -> UserConfiguredModel:
        """Transform a model according to configured rules."""
        model_copy = copy.deepcopy(model)
        for transform_rule in (LowerCaseNamesRule(),):
            model_copy = transform_rule.transform_model(model_copy)

        return model_copy

    @staticmethod
    def post_validation_transform_model(model: UserConfiguredModel) -> UserConfiguredModel:
        """Transform a model according to configured rules."""
        model_copy = copy.deepcopy(model)
        for transform_rule in (
            CreateProxyMeasureRule(),
            BooleanMeasureAggregationRule(),
            CompositeIdentifierExpressionRule(),
        ):
            model_copy = transform_rule.transform_model(model_copy)

        return model_copy



from __future__ import annotations

import datetime
import logging
import threading
from typing import List, Tuple

import pandas as pd

from metricflow.protocols.sql_client import SqlClient
from metricflow.constraints.time_constraint import TimeRangeConstraint
from metricflow.dataflow.sql_table import SqlTable
from metricflow.time.time_constants import ISO8601_PYTHON_FORMAT
from metricflow.time.time_granularity import TimeGranularity

logger = logging.getLogger(__name__)


class TimeSpineSource:
    """Generates a source table containing all timestamps to use for cumulative metrics."""

    _SPINE_TABLE_NAME = "mf_time_spine"

    def __init__(self, sql_client: SqlClient, schema_name: str) -> None:  # noqa
        self._sql_client = sql_client
        self._schema_name = schema_name
        self._verified_spine_table_exists = False
        self._create_table_lock = threading.Lock()

    @property
    def spine_table(self) -> SqlTable:
        """Table containing all dates"""
        return SqlTable(schema_name=self._schema_name, table_name=TimeSpineSource._SPINE_TABLE_NAME)

    @property
    def time_column_name(self) -> str:
        """Name of the column"""
        return "ds"

    @property
    def time_column_granularity(self) -> TimeGranularity:  # noqa: D
        return TimeGranularity.DAY

    def create_if_necessary(self) -> None:
        """Creates the spine table if necessary."""
        logger.info("Waiting to get the lock for the time spine table")
        with self._create_table_lock:
            logger.info("Got the lock for the time spine table")
            if self._verified_spine_table_exists:
                return
            logger.info(f"Checking if the spine table {self.spine_table.sql} exists")
            if self._sql_client.table_exists(self.spine_table):
                logger.info(f"Spine table {self.spine_table.sql} exists")
                self._verified_spine_table_exists = True
                return
            logger.info(f"Spine table {self.spine_table.sql} does not exist")
            start_date = TimeRangeConstraint.ALL_TIME_BEGIN()
            end_date = TimeRangeConstraint.ALL_TIME_END()

            current_date = start_date
            # Using a union type throws a type error for some reason, so going with this approach
            date_spine_table_datetime_data: List[Tuple[datetime.datetime]] = []
            date_spine_table_str_data: List[Tuple[str]] = []

            if self._sql_client.sql_engine_attributes.timestamp_type_supported:
                while current_date <= end_date:
                    date_spine_table_datetime_data.append((current_date,))
                    current_date = current_date + datetime.timedelta(days=1)
            else:
                while current_date <= end_date:
                    date_spine_table_str_data.append((current_date.strftime(ISO8601_PYTHON_FORMAT),))
                    current_date = current_date + datetime.timedelta(days=1)

            self._sql_client.drop_table(self.spine_table)
            num_rows = (
                len(date_spine_table_datetime_data)
                if date_spine_table_datetime_data
                else len(date_spine_table_str_data)
            )

            logger.info(f"Creating date spine table {self.spine_table.sql} with {num_rows} rows")
            self._sql_client.create_table_from_dataframe(
                sql_table=self.spine_table,
                df=pd.DataFrame(
                    columns=[self.time_column_name],
                    data=date_spine_table_datetime_data or date_spine_table_str_data,
                ),
                chunk_size=1000,
            )
            logger.info(f"Created date spine table {self.spine_table.sql}")
            self._verified_spine_table_exists = True



import datetime

import pytest

from metricflow.constraints.time_constraint import TimeRangeConstraint
from metricflow.model.semantic_model import SemanticModel
from metricflow.specs import MetricSpec, TimeDimensionSpec, TimeDimensionReference
from metricflow.time.time_granularity import TimeGranularity
from metricflow.time.time_granularity_solver import (
    TimeGranularitySolver,
    PartialTimeDimensionSpec,
    RequestTimeGranularityException,
)


@pytest.fixture(scope="session")
def time_granularity_solver(  # noqa: D
    extended_date_semantic_model: SemanticModel,
) -> TimeGranularitySolver:
    return TimeGranularitySolver(extended_date_semantic_model)


def test_validate_day_granuarity_for_day_metric(time_granularity_solver: TimeGranularitySolver) -> None:  # noqa: D
    time_granularity_solver.validate_time_granularity(
        metric_specs=[MetricSpec(element_name="bookings")],
        time_dimension_specs=[
            TimeDimensionSpec(element_name="ds", identifier_links=(), time_granularity=TimeGranularity.DAY)
        ],
    )


def test_validate_month_granuarity_for_day_metric(time_granularity_solver: TimeGranularitySolver) -> None:  # noqa: D
    time_granularity_solver.validate_time_granularity(
        metric_specs=[MetricSpec(element_name="bookings")],
        time_dimension_specs=[
            TimeDimensionSpec(element_name="ds", identifier_links=(), time_granularity=TimeGranularity.MONTH)
        ],
    )


def test_validate_month_granuarity_for_month_metric(time_granularity_solver: TimeGranularitySolver) -> None:  # noqa: D
    time_granularity_solver.validate_time_granularity(
        metric_specs=[MetricSpec(element_name="bookings_monthly")],
        time_dimension_specs=[
            TimeDimensionSpec(element_name="ds", identifier_links=(), time_granularity=TimeGranularity.MONTH)
        ],
    )


def test_validate_month_granuarity_for_day_and_month_metrics(  # noqa: D
    time_granularity_solver: TimeGranularitySolver,
) -> None:
    time_granularity_solver.validate_time_granularity(
        metric_specs=[MetricSpec(element_name="bookings"), MetricSpec(element_name="bookings_monthly")],
        time_dimension_specs=[
            TimeDimensionSpec(element_name="ds", identifier_links=(), time_granularity=TimeGranularity.MONTH)
        ],
    )


def test_validate_year_granularity_for_day_and_month_metrics(  # noqa: D
    time_granularity_solver: TimeGranularitySolver,
) -> None:
    time_granularity_solver.validate_time_granularity(
        metric_specs=[MetricSpec(element_name="bookings"), MetricSpec(element_name="bookings_monthly")],
        time_dimension_specs=[
            TimeDimensionSpec(element_name="ds", identifier_links=(), time_granularity=TimeGranularity.YEAR)
        ],
    )


def test_validate_day_granuarity_for_month_metric(time_granularity_solver: TimeGranularitySolver) -> None:  # noqa: D
    with pytest.raises(RequestTimeGranularityException):
        time_granularity_solver.validate_time_granularity(
            metric_specs=[MetricSpec(element_name="bookings_monthly")],
            time_dimension_specs=[
                TimeDimensionSpec(element_name="ds", identifier_links=(), time_granularity=TimeGranularity.DAY)
            ],
        )


def test_validate_day_granularity_for_day_and_month_metric(  # noqa: D
    time_granularity_solver: TimeGranularitySolver,
) -> None:
    with pytest.raises(RequestTimeGranularityException):
        time_granularity_solver.validate_time_granularity(
            metric_specs=[MetricSpec(element_name="bookings"), MetricSpec(element_name="bookings_monthly")],
            time_dimension_specs=[
                TimeDimensionSpec(element_name="ds", identifier_links=(), time_granularity=TimeGranularity.DAY)
            ],
        )


def test_granularity_solution_for_day_metric(time_granularity_solver: TimeGranularitySolver) -> None:  # noqa: D
    assert time_granularity_solver.resolve_granularity_for_partial_time_dimension_specs(
        metric_specs=[MetricSpec(element_name="bookings")],
        partial_time_dimension_specs=[PartialTimeDimensionSpec(element_name="ds", identifier_links=())],
        primary_time_dimension_reference=TimeDimensionReference(element_name="ds"),
    ) == {
        PartialTimeDimensionSpec(element_name="ds", identifier_links=()): TimeDimensionSpec(
            element_name="ds", identifier_links=(), time_granularity=TimeGranularity.DAY
        ),
    }


def test_granularity_solution_for_month_metric(time_granularity_solver: TimeGranularitySolver) -> None:  # noqa: D
    assert time_granularity_solver.resolve_granularity_for_partial_time_dimension_specs(
        metric_specs=[MetricSpec(element_name="bookings_monthly")],
        partial_time_dimension_specs=[PartialTimeDimensionSpec(element_name="ds", identifier_links=())],
        primary_time_dimension_reference=TimeDimensionReference(element_name="ds"),
    ) == {
        PartialTimeDimensionSpec(element_name="ds", identifier_links=()): TimeDimensionSpec(
            element_name="ds", identifier_links=(), time_granularity=TimeGranularity.MONTH
        ),
    }


def test_granularity_solution_for_day_and_month_metrics(  # noqa: D
    time_granularity_solver: TimeGranularitySolver,
) -> None:
    assert time_granularity_solver.resolve_granularity_for_partial_time_dimension_specs(
        metric_specs=[MetricSpec(element_name="bookings"), MetricSpec(element_name="bookings_monthly")],
        partial_time_dimension_specs=[PartialTimeDimensionSpec(element_name="ds", identifier_links=())],
        primary_time_dimension_reference=TimeDimensionReference(element_name="ds"),
    ) == {
        PartialTimeDimensionSpec(element_name="ds", identifier_links=()): TimeDimensionSpec(
            element_name="ds", identifier_links=(), time_granularity=TimeGranularity.MONTH
        )
    }


def test_time_granularity_parameter(  # noqa: D
    time_granularity_solver: TimeGranularitySolver,
) -> None:
    assert time_granularity_solver.resolve_granularity_for_partial_time_dimension_specs(
        metric_specs=[MetricSpec(element_name="bookings"), MetricSpec(element_name="bookings_monthly")],
        partial_time_dimension_specs=[PartialTimeDimensionSpec(element_name="ds", identifier_links=())],
        primary_time_dimension_reference=TimeDimensionReference(element_name="ds"),
        time_granularity=TimeGranularity.YEAR,
    ) == {
        PartialTimeDimensionSpec(element_name="ds", identifier_links=()): TimeDimensionSpec(
            element_name="ds", identifier_links=(), time_granularity=TimeGranularity.YEAR
        )
    }


def test_invalid_time_granularity_parameter(  # noqa: D
    time_granularity_solver: TimeGranularitySolver,
) -> None:
    with pytest.raises(RequestTimeGranularityException, match="Can't use time granularity.*"):
        time_granularity_solver.resolve_granularity_for_partial_time_dimension_specs(
            metric_specs=[MetricSpec(element_name="bookings"), MetricSpec(element_name="bookings_monthly")],
            partial_time_dimension_specs=[PartialTimeDimensionSpec(element_name="ds", identifier_links=())],
            primary_time_dimension_reference=TimeDimensionReference(element_name="ds"),
            time_granularity=TimeGranularity.DAY,
        )


def test_adjusted_time_constraint(time_granularity_solver: TimeGranularitySolver) -> None:  # noqa: D
    assert time_granularity_solver.adjust_time_range_to_granularity(
        time_range_constraint=TimeRangeConstraint(
            start_time=datetime.datetime(2020, 1, 1),
            end_time=datetime.datetime(2020, 1, 15),
        ),
        time_granularity=TimeGranularity.DAY,
    ) == TimeRangeConstraint(
        start_time=datetime.datetime(2020, 1, 1),
        end_time=datetime.datetime(2020, 1, 15),
    )

    assert time_granularity_solver.adjust_time_range_to_granularity(
        time_range_constraint=TimeRangeConstraint(
            start_time=datetime.datetime(2020, 1, 1),
            end_time=datetime.datetime(2020, 1, 15),
        ),
        time_granularity=TimeGranularity.MONTH,
    ) == TimeRangeConstraint(
        start_time=datetime.datetime(2020, 1, 1),
        end_time=datetime.datetime(2020, 1, 31),
    )

    assert time_granularity_solver.adjust_time_range_to_granularity(
        time_range_constraint=TimeRangeConstraint(
            start_time=datetime.datetime(2020, 1, 15),
            end_time=datetime.datetime(2020, 2, 15),
        ),
        time_granularity=TimeGranularity.MONTH,
    ) == TimeRangeConstraint(
        start_time=datetime.datetime(2020, 1, 1),
        end_time=datetime.datetime(2020, 2, 29),
    )


def test_adjusted_time_constraint_for_week_granularity(
    time_granularity_solver: TimeGranularitySolver,
) -> None:
    """Tests the time range adjustment behavior for weekly granularities"""
    assert time_granularity_solver.adjust_time_range_to_granularity(
        time_range_constraint=TimeRangeConstraint(
            start_time=datetime.datetime(2020, 1, 8),
            end_time=datetime.datetime(2020, 1, 15),
        ),
        time_granularity=TimeGranularity.WEEK,
    ) == TimeRangeConstraint(
        start_time=datetime.datetime(2020, 1, 6),
        end_time=datetime.datetime(2020, 1, 19),
    )


def test_adjusted_time_constraint_at_boundaries(
    time_granularity_solver: TimeGranularitySolver,
) -> None:
    """Tests the time range adjustment behavior at the min / max boundaries"""
    assert (
        time_granularity_solver.adjust_time_range_to_granularity(
            time_range_constraint=TimeRangeConstraint(
                start_time=TimeRangeConstraint.all_time().start_time + datetime.timedelta(days=-30),
                end_time=TimeRangeConstraint.all_time().end_time + datetime.timedelta(days=30),
            ),
            time_granularity=TimeGranularity.WEEK,
        )
        == TimeRangeConstraint.all_time()
    )



from __future__ import annotations

import pytest

from metricflow.dataflow.builder.costing import DefaultCostFunction
from metricflow.dataflow.builder.dataflow_plan_builder import DataflowPlanBuilder
from metricflow.model.semantic_model import SemanticModel
from metricflow.plan_conversion.time_spine import TimeSpineSource
from metricflow.protocols.sql_client import SqlClient
from metricflow.specs import TimeDimensionReference
from metricflow.dataset.data_source_adapter import DataSourceDataSet
from metricflow.test.fixtures.model_fixtures import ConsistentIdObjectRepository, multihop_model_data_sets
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.fixtures.sql_client_fixtures import sql_client  # noqa: F401, F403


@pytest.fixture(scope="session")
def composite_dataflow_plan_builder(  # noqa: D
    composite_identifier_semantic_model: SemanticModel,
    consistent_id_object_repository: ConsistentIdObjectRepository,
    time_spine_source: TimeSpineSource,
) -> DataflowPlanBuilder[DataSourceDataSet]:

    return DataflowPlanBuilder(
        data_sets=list(consistent_id_object_repository.composite_model_data_sets.values()),
        semantic_model=composite_identifier_semantic_model,
        primary_time_dimension_reference=TimeDimensionReference(
            element_name="ds",
        ),
        cost_function=DefaultCostFunction[DataSourceDataSet](),
        time_spine_source=time_spine_source,
    )


@pytest.fixture
def dataflow_plan_builder(  # noqa: D
    simple_semantic_model: SemanticModel,
    consistent_id_object_repository: ConsistentIdObjectRepository,
    time_spine_source: TimeSpineSource,
) -> DataflowPlanBuilder[DataSourceDataSet]:
    """Using 'function' scope to make ID generation more deterministic.

    Using 'session' scope can result in other 'session' scope fixtures causing ID consistency issues.
    """
    return DataflowPlanBuilder(
        data_sets=list(consistent_id_object_repository.simple_model_data_sets.values()),
        semantic_model=simple_semantic_model,
        primary_time_dimension_reference=TimeDimensionReference(
            element_name="ds",
        ),
        cost_function=DefaultCostFunction[DataSourceDataSet](),
        time_spine_source=time_spine_source,
    )


@pytest.fixture
def multihop_dataflow_plan_builder(  # noqa: D
    multi_hop_join_semantic_model: SemanticModel,
    time_spine_source: TimeSpineSource,
) -> DataflowPlanBuilder[DataSourceDataSet]:

    return DataflowPlanBuilder(
        data_sets=list(multihop_model_data_sets(multi_hop_join_semantic_model).values()),
        semantic_model=multi_hop_join_semantic_model,
        primary_time_dimension_reference=TimeDimensionReference(
            element_name="ds",
        ),
        cost_function=DefaultCostFunction[DataSourceDataSet](),
        time_spine_source=time_spine_source,
    )


@pytest.fixture(scope="session")
def time_spine_source(  # noqa: D
    sql_client: SqlClient, mf_test_session_state: MetricFlowTestSessionState  # noqa: F811
) -> TimeSpineSource:
    time_spine_source = TimeSpineSource(sql_client=sql_client, schema_name=mf_test_session_state.mf_system_schema)
    time_spine_source.create_if_necessary()
    return time_spine_source



from __future__ import annotations

import logging

from metricflow.instances import (
    InstanceSet,
)

logger = logging.getLogger(__name__)


class DataSet:
    """Describes a set of data that a source node in the dataflow plan contains."""

    def __init__(self, instance_set: InstanceSet) -> None:  # noqa:
        self._instance_set = instance_set

    @property
    def instance_set(self) -> InstanceSet:
        """Returns the instances contained in this dataset."""
        return self._instance_set

    def __repr__(self) -> str:  # noqa: D
        return f"{self.__class__.__name__}()"



from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.validations.dim_names import DimensionAndIdentifierNameValidator


def test_dimension_name_validator(simple_model__pre_transforms: UserConfiguredModel) -> None:  # noqa: D
    dim_name_validator = DimensionAndIdentifierNameValidator(simple_model__pre_transforms)

    # Check a local dimension.
    assert dim_name_validator.is_dimension_valid_for_metric(metric_name="bookings", dimension_name="is_instant")

    # Check a dimension in a different data source that's specified like a local dimension.
    assert not dim_name_validator.is_dimension_valid_for_metric(metric_name="bookings", dimension_name="country_latest")

    # Check dimension in a different data source that can be joined in.
    assert dim_name_validator.is_dimension_valid_for_metric(
        metric_name="bookings", dimension_name="listing__country_latest"
    )

    # Check dimension in a different data source that can't be joined in.
    assert not dim_name_validator.is_dimension_valid_for_metric(
        metric_name="bookings", dimension_name="verification__verification_type"
    )



from __future__ import annotations

from hashlib import sha1
from typing import List, Optional

from metricflow.errors.errors import ParsingException
from metricflow.model.objects.constraints.where import WhereClauseConstraint
from metricflow.model.objects.utils import ParseableObject, ParseableField, HashableBaseModel
from metricflow.object_utils import ExtendedEnum
from metricflow.specs import MeasureReference
from metricflow.time.time_granularity import TimeGranularity
from metricflow.time.time_granularity import string_to_time_granularity


class MetricType(ExtendedEnum):
    """Currently supported metric types"""

    MEASURE_PROXY = "measure_proxy"
    RATIO = "ratio"
    EXPR = "expr"
    CUMULATIVE = "cumulative"


class CumulativeMetricWindow(HashableBaseModel, ParseableField):
    """Describes the window of time the metric should be accumulated over. ie '1 day', '2 weeks', etc"""

    count: int
    granularity: TimeGranularity

    def to_string(self) -> str:  # noqa: D
        return f"{self.count} {self.granularity.value}"

    @staticmethod
    def parse(window: str) -> CumulativeMetricWindow:
        """Returns window values if parsing succeeds, None otherwise

        Output of the form: (<time unit count>, <time granularity>, <error message>) - error message is None if window is formatted properly
        """
        parts = window.split(" ")
        if len(parts) != 2:
            raise ParsingException(
                f"Invalid window ({window}) in cumulative metric. Should be of the form `<count> <granularity>`, ie `28 days`",
            )

        granularity = parts[1]
        # if we switched to python 3.9 this could just be `granularity = parts[0].removesuffix('s')
        if granularity.endswith("s"):
            # months -> month
            granularity = granularity[:-1]
        if granularity not in [item.value for item in TimeGranularity]:
            raise ParsingException(
                f"Invalid time granularity {granularity} in cumulative metric window string: ({window})",
            )

        count = parts[0]
        if not count.isdigit():
            raise ParsingException(f"Invalid count ({count}) in cumulative metric window string: ({window})")

        return CumulativeMetricWindow(
            count=int(count),
            granularity=string_to_time_granularity(granularity),
        )


class MetricTypeParams(HashableBaseModel, ParseableObject):
    """Type params add additional context to certain metric types (the context depends on the metric type)"""

    measure: Optional[MeasureReference]
    measures: Optional[List[MeasureReference]]
    numerator: Optional[MeasureReference]
    denominator: Optional[MeasureReference]
    expr: Optional[str]
    window: Optional[CumulativeMetricWindow]
    grain_to_date: Optional[TimeGranularity]


class Metric(HashableBaseModel, ParseableObject):
    """Describes a metric"""

    name: str
    type: MetricType
    type_params: MetricTypeParams
    constraint: Optional[WhereClauseConstraint]

    @property
    def measure_names(self) -> List[MeasureReference]:  # noqa: D
        tp = self.type_params
        res = tp.measures or []
        if tp.measure:
            res.append(tp.measure)
        if tp.numerator:
            res.append(tp.numerator)
        if tp.denominator:
            res.append(tp.denominator)

        return res

    @property
    def definition_hash(self) -> str:  # noqa: D
        values: List[Optional[str]] = [self.name, self.type_params.expr or ""]
        if self.constraint:
            values.append(self.constraint.where)
            if self.constraint.linkable_names:
                values.extend(self.constraint.linkable_names)
        values.extend([m.element_name for m in self.measure_names])

        hash_builder = sha1()
        for s in values:
            if s is None:
                continue
            hash_builder.update(s.encode("utf-8"))
        return hash_builder.hexdigest()
